"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import argparse
import os
import shutil
import sys
from typing import Optional

import yaml

from . import (
    __version__,
    BuildRunner,
    BuildRunnerConfigurationError,
)
from buildrunner import config, loggers
from buildrunner.config import BuildRunnerConfig
from buildrunner.utils import epoch_time


def parse_args(argv):
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(
        prog=argv[0], description="buildrunner runs builds defined in buildrunner.yaml"
    )

    parser.add_argument(
        "-c",
        "--global-config",
        default=None,
        dest="global_config_file",
        help='global configuration file (defaults to "~/.buildrunner.yaml")',
    )

    parser.add_argument(
        "-d",
        "--directory",
        default=os.getcwd(),
        dest="directory",
        help="build directory (defaults to current working directory)",
    )
    parser.add_argument(
        "-b",
        "--build-results-directory",
        dest="build_results_dir",
        help="build results directory (defaults to <build-directory>/buildrunner.results)",
    )

    parser.add_argument(
        "-f",
        "--file",
        default=None,
        dest="config_file",
        help='build configuration file (defaults to "buildrunner.yaml")',
    )

    parser.add_argument(
        "-x",
        "--debug",
        dest="debug",
        action="store_true",
        help="enables debug logging",
    )

    parser.add_argument(
        "--container-labels",
        default=None,
        help=(
            "arbitrary labels to add to every container started by buildrunner, "
            "in the form of 'key1=value1,key2=value2'"
        ),
    )

    parser.add_argument(
        "-n",
        "--build-number",
        default=None,
        dest="build_number",
        help="build number (defaults to unix/epoch time)",
    )

    parser.add_argument(
        "-t",
        "--docker-timeout",
        default=600,
        type=int,
        dest="docker_timeout",
        help="docker timeout in seconds, defaults to 600",
    )

    parser.add_argument(
        "--push",
        default=False,
        action="store_true",
        dest="push",
        help="push images to remote registries (without this flag buildrunner simply tags images)",
    )

    parser.add_argument(
        "--keep-images",
        default=False,
        action="store_true",
        dest="keep_images",
        help="DEPRECATED: The keep-images argument is deprecated. "
        "Keep images is the default behavior. Please use --cleanup-images instead to cleanup images.",
    )

    parser.add_argument(
        "--cleanup-images",
        default=False,
        action="store_true",
        dest="cleanup_images",
        help="Cleanup generated images at the end of the build",
    )

    parser.add_argument(
        "--local-images",
        default=False,
        action="store_true",
        dest="local_images",
        help="Prefer local images rather than fetching remote images. "
        "This can be used for testing images prior to pushing them and is the equivalent "
        "of setting pull to false for every image/build.",
    )

    parser.add_argument(
        "--platform",
        default=None,
        dest="platform",
        type=str,
        help="The platform architecture to pass to the docker daemon when pulling, building, and running images."
        'For example, "linux/amd64" or "linux/arm64/v8".',
    )

    parser.add_argument(
        "--keep-step-artifacts",
        default=False,
        action="store_true",
        dest="keep_step_artifacts",
        help="keep artifacts generated for each step of the build (step artifacts are by default deleted to "
        "prevent clutter on build machines)",
    )

    parser.add_argument(
        "--clean-cache",
        default=False,
        action="store_true",
        dest="clean_cache",
        help='Clean local caches as defined in buildrunner config ("caches-root"), which defaults to '
        "~/.buildrunner/caches",
    )

    parser.add_argument(
        "-s",
        "--steps",
        default=[],
        dest="steps",
        action="append",
        help="only run the listed steps (use the argument multiple times or specify as comma-delimited)",
    )

    parser.add_argument(
        "--publish-ports",
        default=False,
        action="store_true",
        dest="publish_ports",
        help="publish ports defined on a run step, this should never be used on a build server",
    )

    parser.add_argument(
        "--disable-timestamps",
        default=False,
        action="store_true",
        dest="disable_timestamps",
        help="disables printing of timestamps in the logging output",
    )

    parser.add_argument(
        "--version",
        default=False,
        action="store_true",
        dest="print_version",
        help="print the current buildrunner version and exit",
    )

    parser.add_argument(
        "--no-color",
        default=False,
        action="store_true",
        dest="no_log_color",
        help="disable colors when logging",
    )

    parser.add_argument(
        "--print-generated-files",
        default=False,
        action="store_true",
        dest="print_generated_files",
        help="logs the Jinja generated file contents and stops",
    )

    parser.add_argument(
        "--log-generated-files",
        default=False,
        action="store_true",
        dest="log_generated_files",
        help="logs the Jinja generated file contents",
    )

    parser.add_argument(
        "--disable-multi-platform",
        default=False,
        action="store_true",
        dest="disable_multi_platform",
        help="overrides the 'platforms' configuration and global config; to disable multi-platform builds",
    )

    # Security scan config
    parser.add_argument(
        "--security-scan-enabled",
        default=None,
        choices=["true", "false"],
        dest="security_scan_enabled",
        help="overrides the security-scan.enabled global configuration parameter",
    )
    parser.add_argument(
        "--security-scan-scanner",
        required=False,
        choices=["trivy"],
        dest="security_scan_scanner",
        help="overrides the security-scan.scanner global configuration parameter",
    )
    parser.add_argument(
        "--security-scan-version",
        dest="security_scan_version",
        help="overrides the security-scan.version global configuration parameter",
    )
    parser.add_argument(
        "--security-scan-config-file",
        dest="security_scan_config_file",
        help="overrides the security-scan.config global configuration parameter by loading YAML data from the file",
    )
    parser.add_argument(
        "--security-scan-max-score-threshold",
        type=float,
        dest="security_scan_max_score_threshold",
        help="overrides the security-scan.max-score-threshold global configuration parameter",
    )

    args = parser.parse_args(argv[1:])

    # Set build results dir if not set
    if not args.build_results_dir:
        args.build_results_dir = os.path.join(args.directory, config.RESULTS_DIR)

    # Only absolute directories can do a mount bind
    args.directory = os.path.realpath(args.directory)

    _steps = []
    for _step in args.steps:
        _steps.extend(_step.split(","))
    args.steps = _steps

    return args


def clean_cache():
    """Cache cleanup"""
    args = parse_args(sys.argv)
    BuildRunnerConfig.initialize_instance(
        push=False,
        build_number=1,
        build_id="",
        vcs=None,
        steps_to_run=[],
        build_dir=args.directory,
        global_config_file=args.global_config_file,
        run_config_file=args.config_file,
        build_time=epoch_time(),
        log_generated_files=(
            bool(args.log_generated_files or args.print_generated_files)
        ),
        # Do not attempt to load run configuration, just global configuration
        load_run_config=False,
        global_config_overrides=_get_global_config_overrides(args),
        platform=args.platform,
    )
    BuildRunner.clean_cache()


def _create_results_dir(cleanup_step_artifacts: bool, build_results_dir: str) -> None:
    try:
        # cleanup existing results dir (if needed)
        if cleanup_step_artifacts and os.path.exists(build_results_dir):
            shutil.rmtree(build_results_dir)
        # (re)create the build results dir
        os.makedirs(build_results_dir, exist_ok=True)
    except OSError as exc:
        sys.stderr.write(f"ERROR: {str(exc)}\n")
        sys.exit(os.EX_UNAVAILABLE)


def _load_security_scan_config_file(config_file: Optional[str]) -> Optional[dict]:
    if not config_file:
        return None
    if not os.path.exists(config_file):
        sys.stderr.write(
            f"ERROR: The specified security scan config file ({config_file}) could not be found"
        )
        sys.exit(os.EX_CONFIG)
    try:
        with open(config_file, "r", encoding="utf8") as fobj:
            data = yaml.safe_load(fobj)
            if not data or not isinstance(data, dict):
                sys.stderr.write(
                    f"ERROR: The specified security scan config file ({config_file}) must contain a dictionary"
                )
                sys.exit(os.EX_CONFIG)
            return data
    except Exception as exc:
        sys.stderr.write(
            f"ERROR: The specified security scan config file ({config_file}) could not be loaded: {exc}"
        )
        sys.exit(os.EX_CONFIG)


def _get_security_scan_options(args: argparse.Namespace) -> dict:
    security_scan_config = {
        "enabled": _get_true_value(args.security_scan_enabled),
        "scanner": args.security_scan_scanner,
        "version": args.security_scan_version,
        "config": _load_security_scan_config_file(args.security_scan_config_file),
        "max-score-threshold": args.security_scan_max_score_threshold,
    }
    final_config = {
        key: value for key, value in security_scan_config.items() if value is not None
    }
    if final_config:
        return {"security-scan": final_config}
    return {}


def _get_global_config_overrides(args: argparse.Namespace) -> dict:
    """
    Creates a dictionary of overrides to be deeply merged into the loaded global config file(s) data.
    Note that these field names must match exact what is stored in the global config file(s) and
    that any fields listed here will override configured values. In other words, undefined/none values
    should be filtered from the return data to prevent overriding
    :param args: the parsed CLI args
    :return: the overrides (if any specified)
    """
    overrides = {}
    if args.disable_multi_platform:
        overrides["disable-multi-platform"] = True
    return {
        **overrides,
        **_get_security_scan_options(args),
    }


def _get_true_value(value: Optional[str]) -> Optional[bool]:
    return None if value is None else value == "true"


def initialize_br(args: argparse.Namespace) -> BuildRunner:
    _create_results_dir(not args.keep_step_artifacts, args.build_results_dir)
    loggers.initialize_root_logger(
        args.debug,
        args.no_log_color,
        args.disable_timestamps,
        args.build_results_dir,
    )

    # set build time
    build_time = epoch_time()

    # set build number
    build_number = args.build_number
    if not build_number:
        build_number = build_time

    return BuildRunner(
        build_dir=args.directory,
        build_results_dir=args.build_results_dir,
        global_config_file=args.global_config_file,
        run_config_file=args.config_file,
        build_time=build_time,
        build_number=build_number,
        push=args.push,
        cleanup_images=args.cleanup_images,
        cleanup_cache=args.clean_cache,
        steps_to_run=args.steps,
        publish_ports=args.publish_ports,
        log_generated_files=(
            bool(args.log_generated_files or args.print_generated_files)
        ),
        docker_timeout=args.docker_timeout,
        local_images=args.local_images,
        platform=args.platform,
        global_config_overrides=_get_global_config_overrides(args),
        container_labels=args.container_labels,
    )


def main():
    """Main program execution."""
    args = parse_args(sys.argv)

    # are we just printing the version?
    if args.print_version:
        print(__version__)
        return os.EX_OK

    try:
        build_runner = initialize_br(args)
        if not args.print_generated_files:
            build_runner.run()
            if build_runner.exit_code:
                return build_runner.exit_code
    except BuildRunnerConfigurationError as brce:
        print((str(brce)))
        return os.EX_CONFIG
    return os.EX_OK


if __name__ == "__main__":
    sys.exit(main())



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import os


from buildrunner.docker import DOCKER_DEFAULT_DOCKERD_URL

DAEMON_IMAGE_NAME = "busybox:latest"


class DockerDaemonProxy:
    """
    Class used to encapsulate Docker daemon information within a container.
    """

    def __init__(self, docker_client, log, docker_registry, container_labels, network):
        """ """
        self.docker_client = docker_client
        self.docker_registry = docker_registry
        self.log = log
        self.container_labels = container_labels
        self.network = network
        self._daemon_container = None
        self._env = {
            "DOCKER_HOST": DOCKER_DEFAULT_DOCKERD_URL,
        }

    def get_info(self):
        """
        Return a tuple where the first item is the daemon container id and
        the second is a dict of environment variables to be injected into other
        containers providing settings for docker clients to connect to the
        encapsulated daemon.
        """
        return self._daemon_container, self._env

    def start(self):
        """
        Starts a Docker container encapsulating information to connect to the
        current docker daemon.
        """
        _volumes = []
        _binds = {}

        # setup docker env and mounts so that the docker daemon is accessible
        # from within the run container
        for env_name, env_value in os.environ.items():
            if env_name == "DOCKER_HOST":
                self._env["DOCKER_HOST"] = env_value
            if env_name == "DOCKER_TLS_VERIFY" and env_value:
                self._env["DOCKER_TLS_VERIFY"] = "1"
            if env_name == "DOCKER_CERT_PATH":
                if os.path.exists(env_value):
                    _volumes.append("/dockerdaemon/certs")
                    _binds[env_value] = {
                        "bind": "/dockerdaemon/certs",
                        "ro": True,
                    }
                    self._env["DOCKER_CERT_PATH"] = "/dockerdaemon/certs"

        # if DOCKER_HOST is a unix socket we need to mount the socket in the
        # container and adjust the DOCKER_HOST variable accordingly
        docker_host = self._env["DOCKER_HOST"]
        if docker_host.startswith("unix://"):
            # need to map the socket as a volume
            local_socket = docker_host.replace("unix://", "")
            if os.path.exists(local_socket):
                _volumes.append("/dockerdaemon/docker.sock")
                _binds[local_socket] = {
                    "bind": "/dockerdaemon/docker.sock",
                    "ro": False,
                }
                self._env["DOCKER_HOST"] = "unix:///dockerdaemon/docker.sock"

        # create and start the Docker container
        self._daemon_container = self.docker_client.create_container(
            f"{self.docker_registry}/{DAEMON_IMAGE_NAME}",
            command="/bin/sh",
            volumes=_volumes,
            host_config=self.docker_client.create_host_config(binds=_binds),
            labels=self.container_labels,
            networking_config=self.docker_client.create_networking_config({
                self.network: self.docker_client.create_endpoint_config()
            })
            if self.network
            else None,
        )["Id"]
        self.docker_client.start(self._daemon_container)
        self.log.write(
            f"Created Docker daemon container {self._daemon_container:.10}\n"
        )

    def stop(self):
        """
        Stops the Docker daemon container.
        """
        # kill container
        self.log.write(
            f"Destroying Docker daemon container {self._daemon_container:.10}\n"
        )
        try:
            if self._daemon_container:
                self.docker_client.remove_container(
                    self._daemon_container,
                    force=True,
                    v=True,
                )
        except Exception as _ex:
            self.log.write(f"Failed to remove Docker daemon container: {_ex}\n")



"""
Copyright 2024 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import getpass
import logging
import os
import platform as system_platform
import tempfile
from typing import List, Optional

import vcsinfo

from buildrunner.errors import BuildRunnerConfigurationError
from buildrunner.sshagent import load_ssh_key_from_file, load_ssh_key_from_str
from buildrunner.utils import (
    epoch_time,
    sanitize_tag,
)
from .loader import MASTER_GLOBAL_CONFIG_FILE, load_global_config_files, load_run_file
from .models import (
    generate_and_validate_global_config,
    generate_and_validate_config,
    GlobalConfig,
    Config,
)


DEFAULT_GLOBAL_CONFIG_FILES = [
    MASTER_GLOBAL_CONFIG_FILE,
    "~/.buildrunner.yaml",
]
DEFAULT_RUN_CONFIG_FILES = ["buildrunner.yaml"]
RESULTS_DIR = "buildrunner.results"
LOGGER = logging.getLogger(__name__)


class BuildRunnerConfig:
    """
    Class used to manage buildrunner config.
    """

    CONTEXT_ENV_PREFIXES = [
        "ARTIFACTORY_",
        "BUILDRUNNER_",
        "VCSINFO_",
        "PACKAGER_",
        "GAUNTLET_",
    ]
    PUSH_ENV_VAR_NAME = "BUILDRUNNER_DO_PUSH"

    _INSTANCE: "BuildRunnerConfig" = None

    def __init__(
        self,
        *,
        # These are required for the env
        push: bool,
        build_number: int,
        build_id: str,
        vcs: Optional[vcsinfo.VCS],
        steps_to_run: Optional[List[str]],
        # These are for the config itself
        build_dir: str,
        global_config_file: Optional[str],
        run_config_file: Optional[str],
        log_generated_files: bool,
        build_time: int,
        global_config_overrides: dict,
        platform: Optional[str],
        # Arbitrary labels to add to all started containers, of the form key1=value1,key2=value2
        container_labels: Optional[str] = None,
        # May be passed in to add temporary files to this list as they are created
        tmp_files: Optional[List[str]] = None,
        # Used only from CLI commands that do not need a run config
        load_run_config: bool = True,
    ):  # pylint: disable=too-many-arguments
        self.vcs = vcs
        self.build_dir = build_dir
        self.default_tag = sanitize_tag(build_id)
        self.log_generated_files = log_generated_files
        self.build_time = build_time
        if not self.build_time:
            self.build_time = epoch_time()
        self.container_labels = self._parse_container_labels(container_labels)
        self.tmp_files = tmp_files

        self.global_config = self._load_global_config(
            global_config_file, global_config_overrides
        )
        self.env = self._load_env(
            push,
            build_number=build_number,
            build_id=build_id,
            vcs=vcs,
            steps_to_run=steps_to_run,
            platform=platform,
        )
        self.run_config = (
            self._load_run_config(run_config_file) if load_run_config else None
        )

    @staticmethod
    def _parse_container_labels(container_labels_str: Optional[str]) -> dict:
        container_labels = {}
        if not container_labels_str:
            return container_labels
        for pair in container_labels_str.split(","):
            if "=" not in pair:
                raise BuildRunnerConfigurationError(
                    "Invalid container label format, must be key=value"
                )
            key, value = pair.split("=", 1)
            container_labels[key] = value
        return container_labels

    def _load_global_config(
        self, global_config_file: Optional[str], global_config_overrides: dict
    ) -> GlobalConfig:
        # load global configuration
        gc_files = DEFAULT_GLOBAL_CONFIG_FILES[:]
        gc_files.append(global_config_file or f"{self.build_dir}/.buildrunner.yaml")

        abs_gc_files = self.to_abs_path(gc_files, return_list=True)

        LOGGER.info("")
        LOGGER.info(f"Global configuration is from: {', '.join(abs_gc_files)}")
        global_config, errors = generate_and_validate_global_config(
            **load_global_config_files(
                build_time=self.build_time,
                global_config_files=abs_gc_files,
                global_config_overrides=global_config_overrides,
            )
        )
        if errors:
            errors_str = "\n".join(errors)
            raise BuildRunnerConfigurationError(
                f"Invalid global configuration, {len(errors)} error(s) found:\n{errors_str}"
            )
        return global_config

    def _load_run_config(self, run_config_file: Optional[str]) -> Config:
        _run_config_file = None
        if run_config_file:
            _run_config_file = self.to_abs_path(run_config_file)
        else:
            for name_to_try in DEFAULT_RUN_CONFIG_FILES:
                _to_try = self.to_abs_path(name_to_try)
                if os.path.exists(_to_try):
                    _run_config_file = _to_try
                    LOGGER.info(f"Found run configuration in {name_to_try}")
                    break

        if not _run_config_file or not os.path.exists(_run_config_file):
            raise BuildRunnerConfigurationError("Cannot find build configuration file")

        run_config, errors = generate_and_validate_config(
            **load_run_file(
                global_config=self.global_config,
                build_time=self.build_time,
                env=self.env,
                run_config_file=_run_config_file,
                log_file=self.log_generated_files,
                default_tag=self.default_tag,
            )
        )
        if errors:
            errors_str = "\n".join(errors)
            raise BuildRunnerConfigurationError(
                f"Invalid configuration, {len(errors)} error(s) found:\n{errors_str}"
            )

        return run_config

    def _get_config_context(
        self,
        *,
        contexts: List[dict],
        build_number: int,
        build_id: str,
        vcs: Optional[vcsinfo.VCS],
        steps_to_run: Optional[List[str]],
        platform: Optional[str],
    ) -> dict:
        """
        Generate the Jinja configuration context for substitution

        Args:
          global_env (dict): Env vars to set from the global config file.
          ctx (dict): A dictionary of key/values to be merged over the default, generated context.

        Returns:
          dict: A dictionary of key/values to be substituted
        """

        context = {
            "BUILDRUNNER_ARCH": str(platform or system_platform.machine()),
            "BUILDRUNNER_BUILD_NUMBER": str(build_number),
            "BUILDRUNNER_BUILD_ID": str(build_id),
            "BUILDRUNNER_BUILD_DOCKER_TAG": str(self.default_tag),
            "BUILDRUNNER_BUILD_TIME": str(self.build_time),
            "BUILDRUNNER_STEPS": steps_to_run,
        }
        if vcs:
            context.update({
                "VCSINFO_NAME": str(vcs.name),
                "VCSINFO_BRANCH": str(vcs.branch),
                "VCSINFO_NUMBER": str(vcs.number),
                "VCSINFO_ID": str(vcs.id),
                "VCSINFO_SHORT_ID": str(vcs.id)[:7],
                "VCSINFO_MODIFIED": str(vcs.modified),
                "VCSINFO_RELEASE": str(vcs.release),
            })

        # Add the global env vars before any other context vars
        for cur_context in contexts:
            if cur_context:
                context.update(cur_context)

        for env_name, env_value in os.environ.items():
            for prefix in self.CONTEXT_ENV_PREFIXES:
                if env_name.startswith(prefix):
                    context[env_name] = env_value

        return context

    def _load_env(self, push: bool, **kwargs) -> dict:
        base_context = {}
        if push:
            base_context[self.PUSH_ENV_VAR_NAME] = 1
        env = self._get_config_context(
            contexts=[base_context, self.global_config.env], **kwargs
        )
        # print out env vars
        # pylint: disable=consider-iterating-dictionary
        key_len = max(len(key) for key in env.keys())
        for key in sorted(env.keys()):
            val = env[key]
            LOGGER.debug(f"Environment: {key!s:>{key_len}}: {val}")
        return env

    @classmethod
    def initialize_instance(cls, *args, **kwargs) -> None:
        cls._INSTANCE = cls(*args, **kwargs)

    @classmethod
    def get_instance(cls) -> "BuildRunnerConfig":
        if not cls._INSTANCE:
            raise Exception("Configuration was accessed before initialization")
        return cls._INSTANCE

    def to_abs_path(self, path, return_list=False):
        """
        Convert a path to an absolute path (if it isn't one already).
        """

        paths = path
        if not isinstance(path, list):
            paths = [path]

        for index, _ in enumerate(paths):
            _path = os.path.expanduser(paths[index])
            if os.path.isabs(_path):
                paths[index] = os.path.realpath(_path)
            else:
                paths[index] = os.path.realpath(os.path.join(self.build_dir, _path))
        if return_list:
            return paths
        return paths[0]

    def get_build_server_from_alias(self, host: str) -> str:
        """
        Given a host alias string determine the actual build server host value
        to use by checking the global configuration.
        """
        # if no build servers configuration in global config just return host
        if not self.global_config.build_servers:
            return host

        for _host, _host_aliases in self.global_config.build_servers.items():
            if host in _host_aliases:
                return _host

        return host

    def get_ssh_keys_from_aliases(self, key_aliases: List[str]) -> list:
        """
        Given a list of key aliases return key objects based on keys registered in the global config.
        """
        if not key_aliases:
            return []
        if not self.global_config.ssh_keys:
            raise BuildRunnerConfigurationError(
                "SSH key aliases specified but no 'ssh-keys' configuration in global build runner config"
            )

        _keys = []
        _matched_aliases = []
        for key_info in self.global_config.ssh_keys:
            if not key_info.aliases:
                continue

            _password = None
            _prompt_for_password = False
            if key_info.password:
                _password = key_info.password
            else:
                _prompt_for_password = key_info.prompt_password

            for alias in key_aliases:
                if alias in key_info.aliases:
                    _matched_aliases.append(alias)

                    # Prompt for password if necessary.  Only once per key
                    if _prompt_for_password:
                        _password = getpass.getpass(f"Password for SSH Key ({alias}): ")
                        _prompt_for_password = False

                    if key_info.file:
                        _key_file = os.path.realpath(
                            os.path.expanduser(os.path.expandvars(key_info.file))
                        )

                        _keys.append(load_ssh_key_from_file(_key_file, _password))
                    elif key_info.key:
                        _keys.append(load_ssh_key_from_str(key_info.key, _password))

        for alias in key_aliases:
            if alias not in _matched_aliases:
                raise BuildRunnerConfigurationError(
                    f"Could not find valid SSH key matching alias '{alias}'"
                )

        return _keys

    def get_local_files_from_alias(self, file_alias: str) -> Optional[str]:
        """
        Given a file alias lookup the local file path from the global config.
        """
        if not file_alias:
            return None
        if not self.global_config.local_files:
            LOGGER.info("No 'local-files' configuration in global build runner config")
            return None

        for local_alias, local_file in self.global_config.local_files.items():
            if file_alias == local_alias:
                local_path = os.path.realpath(
                    os.path.expanduser(os.path.expandvars(local_file))
                )
                if os.path.exists(local_path):
                    return local_path

                # This secondary code will create the file in the location specified and put the contents
                # from the config value into the file for usage in a container. This may ONLY be used in the master
                # global config file.
                # pylint: disable=consider-using-with
                _fileobj = tempfile.NamedTemporaryFile(
                    delete=False,
                    dir=self.global_config.temp_dir,
                )
                _fileobj.write(
                    local_file.encode("utf8")
                    if isinstance(local_file, str)
                    else local_file
                )
                tmp_path = os.path.realpath(_fileobj.name)
                _fileobj.close()
                if self.tmp_files is not None:
                    self.tmp_files.append(tmp_path)
                return tmp_path

        return None



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

from typing import List, Optional

from buildrunner.config.models_step import StepPypiPush
from buildrunner.errors import (
    BuildRunnerConfigurationError,
)
from buildrunner.steprunner.tasks import BuildStepRunnerTask


class PypiRepoDefinition:
    """
    Contains the definition for a PyPi push repository.
    """

    def __init__(
        self,
        repository: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
        skip_existing: bool = False,
    ):
        self.repository = repository
        self.username = username
        self.password = password
        self.skip_existing = skip_existing


class PypiPushBuildStepRunnerTask(BuildStepRunnerTask):
    """
    Class used to push the resulting python packages to the given repository.
    """

    def __init__(self, step_runner, pypi_pushes: List[StepPypiPush]):
        super().__init__(step_runner, pypi_pushes[0])

        if not self.step_runner.build_runner.push:
            # Was not invoked with ``--push`` so just skip this.  This avoids twine
            # complaining when the push repository is not configured and the user
            # is not even interested in pushing.
            return

        self._repos = [
            PypiRepoDefinition(
                push.repository,
                push.username,
                push.password,
                push.skip_existing,
            )
            for push in pypi_pushes
        ]

        imported = False
        for repo in self._repos:
            if repo.repository not in self.step_runner.build_runner.pypi_packages:
                if not imported:
                    # Importing here avoids twine dependency when it is unnecessary
                    import twine.settings  # pylint: disable=import-outside-toplevel
                    import twine.exceptions  # pylint: disable=import-outside-toplevel

                    imported = True

                try:
                    if repo.username is not None and repo.password is not None:
                        upload_settings = twine.settings.Settings(
                            repository_url=repo.repository,
                            username=repo.username,
                            password=repo.password,
                            disable_progress_bar=True,
                            skip_existing=repo.skip_existing,
                        )
                    else:
                        upload_settings = twine.settings.Settings(
                            repository_name=repo.repository,
                            disable_progress_bar=True,
                            skip_existing=repo.skip_existing,
                        )
                except twine.exceptions.InvalidConfiguration as twe:
                    raise BuildRunnerConfigurationError(
                        f'Pypi is unable to find an entry for "{repo.repository}" in your .pypirc.\n'
                    ) from twe

                self.step_runner.build_runner.pypi_packages[repo.repository] = {
                    "upload_settings": upload_settings,
                    "packages": [],
                }

    def run(self, context):
        if not self.step_runner.build_runner.push:
            self.step_runner.log.write('Push not requested with "--push": skipping\n')
            return

        self.step_runner.log.write(
            f"Preparing resulting packages for push to {'/'.join(repo.repository for repo in self._repos)}.\n"
        )

        # get python-sdist packages for this step only
        for _artifact, _attributes in self.step_runner.build_runner.artifacts.items():
            if (
                _artifact.startswith(self.step_runner.name + "/")
                and _attributes
                and "type" in _attributes
                and _attributes["type"] in ("python-wheel", "python-sdist")
            ):
                for repo in self._repos:
                    self.step_runner.build_runner.pypi_packages[repo.repository][
                        "packages"
                    ].append(
                        f"{self.step_runner.build_runner.build_results_dir}/{_artifact}"
                    )



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import json
import os
import requests

from buildrunner.errors import BuildRunnerProvisionerError


class SaltProvisioner:
    """
    Provisioner used to apply a salt state defined in the run configuration.
    """

    def __init__(self, sls, console=None):
        self.sls = sls
        self.console = console

    def provision(self, runner):
        """
        Apply the configured salt state, bootstrapping salt if it is not found.
        """
        if self.console:
            self.console.write("Running salt provisioner...\n")

        # see if salt is installed, bootstrap if it isn't
        if runner.run("salt-call -h") != 0:
            # pull bootstrap and run as a script
            bootstrap_response = requests.get(
                "http://bootstrap.saltstack.org", timeout=600
            )
            if bootstrap_response.status_code != 200:
                raise BuildRunnerProvisionerError("Unable to get salt bootstrap")
            if self.console:
                self.console.write(
                    "Cannot detect salt installation--bootstrapping...\n"
                )
            runner.run_script(
                bootstrap_response.text,
                args="-X -n",
                console=self.console,
            )
            # we don't rely on the bootstrap return code because the
            # ubuntu/debian install fails because the services aren't
            # registered. just rely on salt-call being present instead
            exit_code = runner.run("salt-call -h")
            if exit_code != 0:
                raise BuildRunnerProvisionerError("Unable to bootstrap salt")

        # create tmp file_root dir and write top.sls and dr.sls there
        file_root_dir = runner.tempfile(suffix="_salt_file_root")
        runner.run(f"mkdir -p {file_root_dir}")
        runner.write_to_container_file(
            'base: {"*": ["dr"]}',
            os.path.join(file_root_dir, "top.sls"),
        )
        runner.write_to_container_file(
            json.dumps(dict(self.sls)),
            os.path.join(file_root_dir, "dr.sls"),
        )

        # run a salt-call highstate with the new file_root dir
        salt_call_prefix = ""
        exit_code = runner.run("sudo -h")
        if exit_code == 0:
            salt_call_prefix = "sudo "
        exit_code = runner.run(
            f"{salt_call_prefix}salt-call --local --file-root={file_root_dir} state.highstate",
            console=self.console,
        )
        if exit_code != 0:
            raise BuildRunnerProvisionerError("Unable to provision with salt")



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""


class BuildRunnerError(Exception):
    """Base BuildRunner Exception"""

    pass


class BuildRunnerProtocolError(BuildRunnerError):
    """Error with unhandled state in protocol"""

    pass


class BuildRunnerConfigurationError(BuildRunnerError):
    """Error indicating an issue with the build configuration"""

    pass


class BuildRunnerProcessingError(BuildRunnerError):
    """Error indicating the build should be 'failed'"""

    pass


class BuildRunnerProvisionerError(BuildRunnerError):
    """Error indicating an issue with a provisioner"""

    pass


class BuildRunnerVersionError(Exception):
    """
    Exception class for invalid buildrunner version
    """

    pass


class ConfigVersionFormatError(Exception):
    """
    Exception class for invalid config version format
    """

    pass


class ConfigVersionTypeError(Exception):
    """
    Exception class for invalid config version type
    """

    pass



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import datetime
import io
import os.path
import platform
import socket
import ssl
from collections import OrderedDict
from os import listdir
from os.path import isfile, join, getmtime
from pathlib import Path
import tarfile
from types import GeneratorType
from typing import Optional, Union

from docker.utils import compare_version
from retry import retry
import docker.errors
import timeout_decorator

from buildrunner import BuildRunnerConfig
from buildrunner.docker import (
    new_client,
    force_remove_container,
    BuildRunnerContainerError,
)
from buildrunner.loggers import ConsoleLogger, ContainerLogger, DockerPullProgress
from buildrunner.utils import (
    acquire_flock_open_read_binary,
    acquire_flock_open_write_binary,
    release_flock,
    tempfile,
)

CACHE_NUM_RETRIES = 2
CACHE_TIMEOUT_SECONDS = 240


class BuildRunnerCacheTimeout(Exception):
    """
    Exception which is raised when there is a timeout issue related to caching.
    """

    pass


class BuildRunnerSavingCache(Exception):
    """
    Exception which is raised when there is an issue related to saving cache.
    """

    pass


class DockerRunner:
    """
    An object that manages and orchestrates the lifecycle and execution of a
    Docker container.
    """

    class ImageConfig:
        """
        An object that captures image-specific configuration
        """

        def __init__(self, image_name, pull_image=True, platform=None):
            self.image_name = image_name
            self.pull_image = pull_image
            self.platform = platform

    def __init__(
        self,
        image_config,
        dockerd_url=None,
        log: Union[ContainerLogger, None] = None,
        run_log_debug: bool = False,
    ):
        image_name = image_config.image_name
        pull_image = image_config.pull_image
        image_platform = image_config.platform

        self.image_name = image_name.lower()
        self.platform = image_platform
        if log and self.image_name != image_name:
            log.write(
                f"Forcing image_name to lowercase: {image_name} => {self.image_name}\n"
            )
        self.docker_client = new_client(
            dockerd_url=dockerd_url,
            # Disable timeouts for running commands
            timeout=0,
        )
        self.run_log_debug = run_log_debug
        self.container = None
        self.shell = None
        self.committed_image = None
        self.containers = []

        # By default, pull the image.  If the pull_image parameter is
        # set to False, only pull the image if it can't be found locally
        #
        # Pull all images to ensure we get the hashes for intermediate images
        found_image = False
        for image in self.docker_client.images(all=True):
            if (
                image["Id"].startswith("sha256:" + self.image_name)
                or image["Id"] == self.image_name
            ):
                # If the image name is simply a hash, it refers to an intermediate
                # or imported image.  We don't want to "pull" these, as the hash
                # won't exist as a valid upstream repoistory/image
                found_image = True
                pull_image = False
            else:
                for tag in image["RepoTags"] or []:
                    if tag == self.image_name:
                        found_image = True
                        break
            if found_image:
                # No need to continue once we've found the image
                break

        if pull_image or not found_image:
            if log:
                log.write(f"Pulling image {self.image_name}\n")
            with DockerPullProgress() as docker_progress:
                for data in self.docker_client.pull(
                    self.image_name, stream=True, decode=True, platform=self.platform
                ):
                    docker_progress.status_report(data)
            if log:
                log.write("\nImage pulled successfully\n")

    def _run_log(
        self,
        log: Union[ConsoleLogger, ContainerLogger, None],
        output: Union[str, bytes],
    ):
        """
        Log a message to the given log stream at a debug or info level depending on configuration.
        """
        if log:
            for line in log.clean_output(output).split("\n"):
                if self.run_log_debug:
                    log.debug(line)
                else:
                    log.info(line)

    def start(
        self,
        shell="/bin/sh",
        entrypoint=None,
        working_dir=None,
        name=None,
        volumes=None,
        volumes_from=None,
        links=None,
        ports=None,
        provisioners=None,
        environment=None,
        user=None,
        hostname=None,
        dns=None,
        dns_search=None,
        extra_hosts=None,
        containers=None,
        systemd: bool = False,
        systemd_cgroup2: bool = False,
        cap_add=None,
        privileged=False,
        network=None,
    ):
        """
        Kwargs:
          volumes (dict): mount the local dir (key) to the given container
                          path (value)
        """
        if self.container:
            raise BuildRunnerContainerError("Container already started")
        self.shell = shell

        # save any spawned containers
        if containers:
            self.containers = containers

        # prepare volumes
        _volumes = []
        _binds = {}

        security_opt = None
        command = shell
        tmpfs = {}
        cgroupns = None
        if systemd:
            # If we are running in a systemd context, the following 3 settings are necessary to
            # allow services to run.
            if systemd_cgroup2:
                # Ensure that cgroup v2 is supported before attempting to use it
                # Note: this check only works on linux systems
                if platform.system() == "Linux" and not os.path.exists(
                    "/sys/fs/cgroup/cgroup.controllers"
                ):
                    raise BuildRunnerContainerError(
                        "cgroup v2 is not enabled on this host but is set on the container, please check configuration"
                    )
                volumes["/sys/fs/cgroup/buildrunner.scope"] = "/sys/fs/cgroup:rw"
                tmpfs["/run"] = ""
                cgroupns = "host"
            else:
                volumes["/sys/fs/cgroup"] = "/sys/fs/cgroup:ro"
            security_opt = ["seccomp=unconfined"]
            command = "/usr/sbin/init"

        if volumes:
            for key, value in volumes.items():
                to_bind = value
                _ro = False
                if to_bind.rfind(":") > 0:
                    tokens = to_bind.rsplit(":", 1)
                    to_bind = tokens[0]
                    _ro = tokens[1] == "ro"
                _volumes.append(to_bind)
                _binds[key] = {
                    "bind": to_bind,
                    "ro": _ro,
                }

        # prepare ports
        _port_list = None
        if ports:
            _port_list = list(ports.keys())

        # check args
        if dns_search and isinstance(dns_search, str):
            dns_search = dns_search.split(",")

        kwargs = {
            "name": name,
            "command": command,
            "volumes": _volumes,
            "ports": _port_list,
            "stdin_open": True,
            "tty": True,
            "environment": environment,
            "user": user,
            "working_dir": working_dir,
            "hostname": hostname,
            "labels": BuildRunnerConfig.get_instance().container_labels,
            "host_config": self.docker_client.create_host_config(
                binds=_binds,
                links=links,
                port_bindings=ports,
                volumes_from=volumes_from,
                dns=dns,
                dns_search=dns_search,
                extra_hosts=extra_hosts,
                security_opt=security_opt,
                cap_add=cap_add,
                privileged=privileged,
                tmpfs=tmpfs,
                cgroupns=cgroupns,
            ),
        }
        if entrypoint:
            kwargs["entrypoint"] = entrypoint
            del kwargs["command"]
        if network:
            kwargs["networking_config"] = (
                self.docker_client.create_networking_config({
                    network: self.docker_client.create_endpoint_config(
                        aliases=[hostname] if hostname else None
                    )
                })
                if network
                else None
            )

        if compare_version("1.10", self.docker_client.api_version) < 0:
            kwargs["dns"] = dns

        # start the container
        self.container = self.docker_client.create_container(self.image_name, **kwargs)
        self.docker_client.start(self.container["Id"])

        # run any supplied provisioners
        if provisioners:
            for provisioner in provisioners:
                try:
                    provisioner.provision(self)
                except Exception as ex:
                    self.cleanup()
                    raise ex

        return self.container["Id"]

    def stop(self):
        """
        Stop the backing Docker container.
        """
        if self.container:
            self.docker_client.stop(
                self.container["Id"],
                timeout=0,
            )

    def cleanup(self):
        """
        Cleanup the backing Docker container, stopping it if necessary.
        """
        if self.container:
            for container in self.containers:
                try:
                    force_remove_container(self.docker_client, container)
                except docker.errors.NotFound:
                    try:
                        container_ids = self.docker_client.containers(
                            filters={"label": container}, quiet=True
                        )
                        if container_ids:
                            for container_id in container_ids:
                                try:
                                    self.docker_client.remove_container(
                                        container_id["Id"],
                                        force=True,
                                        v=True,
                                    )
                                except Exception as _ex:
                                    print(
                                        f'Unable to delete docker container with id "{container_id["Id"]}"'
                                    )
                        else:
                            print(
                                f'Unable to find docker container with name or label "{container}"'
                            )
                    except docker.errors.NotFound:
                        print(
                            f'Unable to find docker container with name or label "{container}"'
                        )
            try:
                self.docker_client.remove_container(
                    self.container["Id"],
                    force=True,
                    v=True,
                )
            except docker.errors.NotFound:
                print(
                    f'Unable to delete docker container with id "{self.container["Id"]}"'
                )

        self.container = None

    @staticmethod
    def _get_cache_file_from_prefix(
        logger: ContainerLogger, local_cache_archive_file: str, docker_path: str
    ) -> Optional[str]:
        if os.path.exists(local_cache_archive_file):
            logger.info(
                f"Using cache {local_cache_archive_file} for destination path {docker_path}"
            )
            return local_cache_archive_file
        cache_dir = os.path.dirname(local_cache_archive_file)

        if not os.path.exists(cache_dir):
            logger.info(
                f"Cache directory {cache_dir} does not exist, "
                f"skipping restore of archive {local_cache_archive_file}"
            )
            return None

        files = [f for f in listdir(cache_dir) if isfile(join(cache_dir, f))]

        cache_key = Path(local_cache_archive_file).stem

        most_recent_time = 0
        local_cache_archive_match = None
        for file_name in files:
            if file_name.startswith(cache_key):
                curr_archive_file = join(cache_dir, file_name)
                mod_time = getmtime(curr_archive_file)
                if mod_time > most_recent_time:
                    most_recent_time = mod_time
                    local_cache_archive_match = curr_archive_file

        if local_cache_archive_match is None:
            logger.info(
                f"Not able to restore cache {docker_path} since "
                f"there was no matching prefix for `{local_cache_archive_file}`"
            )
            return None
        logger.info(
            f"Found cache {local_cache_archive_match} matching prefix {local_cache_archive_file} "
            f"for destination path {docker_path}"
        )

        return local_cache_archive_match

    @retry(exceptions=BuildRunnerCacheTimeout, tries=CACHE_NUM_RETRIES)
    @timeout_decorator.timeout(
        seconds=CACHE_TIMEOUT_SECONDS, timeout_exception=BuildRunnerCacheTimeout
    )
    def _put_cache_in_container(self, docker_path: str, file_obj: io.IOBase) -> bool:
        """
        Insert a file or folder in an existing container using a tar archive as
        source.

        :param docker_path: Path of file or folder in the container
        :param file_obj: Opened file object of cache
        :return: True if the call succeeds.
        """
        return self.docker_client.put_archive(
            self.container["Id"], docker_path, file_obj
        )

    def restore_caches(self, logger: ContainerLogger, caches: OrderedDict) -> None:
        """
        Restores caches from the host system to the destination location in the docker container.
        """
        if caches is None or not isinstance(caches, OrderedDict):
            raise TypeError(
                f"Caches should be type OrderDict instead of {type(caches)}"
            )

        restored_cache_src = set()
        for local_cache_archive_file, docker_path in caches.items():
            if docker_path in restored_cache_src:
                logger.info(
                    f"Cache for destination path {docker_path} has already been matched and restored to the container, "
                    f"skipping {local_cache_archive_file}"
                )
                continue

            # Check for prefix matching
            actual_cache_archive_file = self._get_cache_file_from_prefix(
                logger, local_cache_archive_file, docker_path
            )
            if actual_cache_archive_file is None:
                # Errors are printed out in the other method
                continue

            orig_shell = self.shell
            file_obj = None
            try:
                self.shell = "/bin/sh"
                exit_code = self.run(f"mkdir -p {docker_path}")
                if exit_code:
                    logger.warning(
                        f"There was an issue creating {docker_path} on the docker container"
                    )

                # Allow multiple people to read from the file at the same time
                # If another instance adds an exclusive lock in the save method below,
                # this will block until the lock is released
                file_obj = acquire_flock_open_read_binary(
                    lock_file=actual_cache_archive_file, logger=logger
                )
                logger.info(
                    "File lock acquired. Attempting to put cache into the container."
                )

                restored_cache_src.add(docker_path)
                if not self._put_cache_in_container(docker_path, file_obj):
                    logger.warning(
                        f"An error occurred when trying to use cache "
                        f"{actual_cache_archive_file} at the path {docker_path}"
                    )

            except docker.errors.APIError:
                logger.exception("Encountered exception")
            finally:
                self.shell = orig_shell
                release_flock(file_obj, logger)
                logger.info("Cache was put into the container. Released file lock.")

    @retry(exceptions=BuildRunnerCacheTimeout, tries=CACHE_NUM_RETRIES)
    @timeout_decorator.timeout(
        seconds=CACHE_TIMEOUT_SECONDS, timeout_exception=BuildRunnerCacheTimeout
    )
    def _write_cache(self, docker_path: str, file_obj: io.IOBase):
        """
        Write cache locally from file or folder in a running container

        :param docker_path: Path of file or folder in the container
        :param file_obj: Opened file object to write cache
        """
        bits, _ = self.docker_client.get_archive(
            self.container["Id"], f"{docker_path}/."
        )
        for chunk in bits:
            file_obj.write(chunk)

    def write_cache_history_log(
        self,
        log_str: str,
        cache_location: str,
        logger: ContainerLogger,
    ) -> None:
        """
        Writes the cache log to a file in the cache location.

        :param log_str: The log string to write to the file
        :param cache_location: The location of the cache file
        :param logger: The logger to write log messages to
        """

        cache_history_log = f"{cache_location}/cache_history.log"
        file_obj = None
        try:
            file_obj = acquire_flock_open_write_binary(
                lock_file=cache_history_log, logger=logger, mode="a"
            )
            logger.info(
                f"File lock acquired. Attempting to write cache history log to {cache_history_log}"
            )
            file_obj.write(log_str)
        finally:
            release_flock(file_obj, logger)
            logger.info("Writing to cache history log completed, released file lock")

    def save_caches(
        self, logger: ContainerLogger, caches: OrderedDict, env_vars: dict = dict()
    ) -> None:
        """
        Saves caches from a source locations in the docker container to locations on the host system as archive file.
        """
        saved_cache_src = set()
        if caches and isinstance(caches, OrderedDict):
            for local_cache_archive_file, docker_path in caches.items():
                if docker_path not in saved_cache_src:
                    saved_cache_src.add(docker_path)
                    logger.info(
                        f"Saving cache {docker_path} "
                        f"running on container {self.container['Id']} "
                        f"to local cache {local_cache_archive_file}"
                    )

                    log_line = (
                        f"{datetime.datetime.now().strftime('%m/%d/%Y %H:%M:%S')} - "
                        f'The cache file "{local_cache_archive_file}" '
                        f'was written by step "{env_vars.get("BUILDRUNNER_STEP_NAME")}" in '
                        f'"{env_vars.get("VCSINFO_NAME")}:{env_vars.get("VCSINFO_BRANCH")}:{env_vars.get("VCSINFO_SHORT_ID")}" '
                        f"on host "
                        f'"{socket.gethostname()}" [{env_vars.get("BUILDRUNNER_ARCH")}] '
                        "\n"
                    )

                    self.write_cache_history_log(
                        log_line, os.path.dirname(local_cache_archive_file), logger
                    )

                    file_obj = None
                    try:
                        file_obj = acquire_flock_open_write_binary(
                            lock_file=local_cache_archive_file, logger=logger
                        )
                        logger.info("File lock acquired. Attempting to write to cache.")
                        self._write_cache(docker_path, file_obj)
                    except Exception as e:
                        raise BuildRunnerSavingCache(
                            f"There was an error saving cache to {local_cache_archive_file}.\nException: {e}"
                        )
                    finally:
                        release_flock(file_obj, logger)
                        assert tarfile.is_tarfile(local_cache_archive_file), (
                            f"Failed to create cache {local_cache_archive_file} tar file."
                        )
                        logger.info("Writing to cache completed. Released file lock.")
                else:
                    logger.info(
                        f"The following `{docker_path}` in docker has already been saved. "
                        f"It will not be saved again to `{local_cache_archive_file}`"
                    )

    def run(self, cmd, console=None, stream=True, log=None, workdir=None):
        """
        Run the given command in the container.
        """
        # Unused variable
        _ = workdir

        if isinstance(cmd, str):
            cmdv = [self.shell, "-xc", cmd]
        elif (
            hasattr(cmd, "next")
            or hasattr(cmd, "__next__")
            or hasattr(cmd, "__iter__")
            or isinstance(cmd, GeneratorType)
        ):
            cmdv = cmd
        else:
            raise TypeError(f"Unhandled command type: {type(cmd)}:{cmd}")
        # if console is None:
        #    raise Exception('No console!')
        if not self.container:
            raise BuildRunnerContainerError("Container has not been started")
        if not self.shell:
            raise BuildRunnerContainerError(
                "Cannot call run if container cmd not shell"
            )

        self._run_log(log, f"Executing: {cmdv}")

        create_res = self.docker_client.exec_create(
            self.container["Id"],
            cmdv,
            tty=False,
            # workdir=workdir,
        )
        output_buffer = self.docker_client.exec_start(
            create_res,
            stream=stream,
        )
        if isinstance(output_buffer, (bytes, str)):
            self._run_log(console, output_buffer)
            self._run_log(log, output_buffer)
        elif hasattr(output_buffer, "next") or isinstance(output_buffer, GeneratorType):
            try:
                # Buffer chunks into complete lines
                line_buffer = b""
                for chunk in output_buffer:
                    if isinstance(chunk, str):
                        chunk = chunk.encode("utf-8")
                    line_buffer += chunk

                    # Process complete lines
                    while b"\n" in line_buffer:
                        line, line_buffer = line_buffer.split(b"\n", 1)
                        if (
                            line or line_buffer
                        ):  # Don't log empty lines unless there's more content
                            line_with_newline = line + b"\n"
                            self._run_log(console, line_with_newline)
                            self._run_log(log, line_with_newline)

                # Process any remaining content in buffer
                if line_buffer:
                    self._run_log(console, line_buffer)
                    self._run_log(log, line_buffer)
            except socket.timeout:
                # Ignore timeouts since we check for the exit code anyways at the end
                pass
        else:
            warning = f"WARNING: Unexpected output object: {output_buffer}"
            if console:
                console.write(warning)
            if log:
                log.write(warning)
        inspect_res = self.docker_client.exec_inspect(create_res)
        if "ExitCode" in inspect_res:
            if inspect_res["ExitCode"] is None:
                raise BuildRunnerContainerError(
                    f"Error running cmd ({cmd}): exit code is None"
                )
            return inspect_res["ExitCode"]
        raise BuildRunnerContainerError("Error running cmd: no exit code")

    def run_script(
        self,
        script,
        args="",
        console=None,
    ):
        """
        Run the given script within the container.
        """
        container_script_path = self.write_to_container_file(script, console)

        # execute the script
        return self.run(
            f"{container_script_path} {args}",
            console=console,
        )

    def write_to_container_file(self, source_file, logger):
        """
        Writes contents to the given path within the container.
        """
        container_dir = "/tmp"
        container_script_path = f"{container_dir}/{os.path.basename(source_file)}"
        my_tar_file_path = tempfile(prefix="buildrunner-", suffix=".tar")

        try:
            # Create destination directory in container
            self.run(f"mkdir -p {container_dir}", console=logger)

            # Create tar file with script to put in container
            with tarfile.open(my_tar_file_path, "w") as tar:
                tar.add(source_file, arcname=os.path.basename(source_file))

            # Put the script file in the container
            with open(my_tar_file_path, "rb") as tar:
                code = self.docker_client.put_archive(
                    self.container["Id"],
                    os.path.dirname(container_script_path),
                    tar.read(),
                )
                if code:
                    # Make the script executable in the container
                    self.run(f"chmod +x {container_script_path}", console=logger)
                else:
                    logger.error(
                        f"There was an issue putting the script in the container. Code: {code}"
                    )

        except Exception as e:
            raise BuildRunnerContainerError(f"Error writing script to container: {e}")
        finally:
            if os.path.exists(my_tar_file_path):
                os.remove(my_tar_file_path)
            return container_script_path

    def _get_status(self):
        """
        Return the status dict for the container.
        """
        status = None
        try:
            status = self.docker_client.inspect_container(
                self.container["Id"],
            )
        except docker.errors.APIError:
            pass
        return status

    def get_ip(self, network=None):
        """
        Return the ip address of the running container
        """
        ipaddr = None
        try:
            if self.is_running():
                inspection = self.docker_client.inspect_container(
                    self.container["Id"],
                )
                network_settings = inspection.get("NetworkSettings", {})
                if network:
                    ipaddr = (
                        network_settings.get("Networks", {})
                        .get(network, {})
                        .get("IPAddress", None)
                    )
                else:
                    ipaddr = network_settings.get("IPAddress", None)
        except docker.errors.APIError:
            pass
        return ipaddr

    def is_running(self):
        """
        Return whether the container backed by this Runner is currently
        running.
        """
        status = self._get_status()
        if not status:
            return False
        if "State" not in status or "Running" not in status["State"]:
            return False
        return status["State"]["Running"]

    @property
    def exit_code(self):
        """
        Return the exit code of the completed container, or None if it is still
        running.
        """
        status = self._get_status()
        if not status:
            return None
        if "State" not in status or "ExitCode" not in status["State"]:
            return None
        return status["State"]["ExitCode"]

    def attach_until_finished(self, stream=None):
        """
        Attach to the container, writing output to the given log stream until
        the container exits.
        """
        docker_socket: socket.SocketIO = self.docker_client.attach_socket(
            self.container["Id"],
        )
        running = self.is_running()
        while running:
            try:
                for line in docker_socket:
                    if stream:
                        stream.write(line)
            except socket.timeout:
                pass
            except ssl.SSLError as ssle:
                if "The read operation timed out" not in str(ssle):
                    raise
            running = self.is_running()

    def commit(self, stream):
        """
        Commit the ending state of the container as an image, returning the
        image id.
        """
        if self.committed_image:
            return self.committed_image
        if not self.container:
            raise BuildRunnerContainerError("Container not started")
        if self.is_running():
            raise BuildRunnerContainerError("Container is still running")
        stream.write(
            f"Committing build container {self.container['Id']:.10} as an image...\n"
        )
        self.committed_image = self.docker_client.commit(
            self.container["Id"],
        )["Id"]
        stream.write(f"Resulting build container image: {self.committed_image:.10}\n")
        return self.committed_image



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""
# pylint: disable=too-many-lines

from collections import OrderedDict
import grp
import os
import pwd
import threading
import time
import uuid

import python_on_whales

import buildrunner.docker
from buildrunner.config import BuildRunnerConfig
from buildrunner.config.models_step import RunAndServicesBase, StepRun, Service
from buildrunner.docker.daemon import DockerDaemonProxy
from buildrunner.docker.runner import DockerRunner
from buildrunner.errors import (
    BuildRunnerConfigurationError,
    BuildRunnerProcessingError,
)
from buildrunner.provisioners import create_provisioners
from buildrunner.sshagent import DockerSSHAgentProxy
from buildrunner.steprunner.tasks import BuildStepRunnerTask
from buildrunner.steprunner.tasks.build import BuildBuildStepRunnerTask
from buildrunner.loggers import ContainerLogger

DEFAULT_SHELL = "/bin/sh"
SOURCE_VOLUME_MOUNT = "/source"
ARTIFACTS_VOLUME_MOUNT = "/artifacts"
FILE_INFO_DELIMITER = "~!~"


class RunBuildStepRunnerTask(BuildStepRunnerTask):
    """
    Class used to manage "run" build tasks.
    """

    # Lightweight docker image for artifact management
    ARTIFACT_LISTER_DOCKER_IMAGE = "ubuntu:19.04"
    # Lightweight docker image for running nc
    NC_DOCKER_IMAGE = "subfuzion/netcat:latest"

    # Default to 10 min when waiting for ports to be opened
    WAIT_FOR_DEFAULT_TIMEOUT = 600

    TAR_COMPRESSION_ARG = {
        "gz": "--gzip",
        "bz2": "--bzip2",
        "xz": "--xz",
        "lzma": "--lzma",
        "lzip": "--lzip",
        "lzop": "--lzop",
        "z": "-Z",
    }

    def __init__(self, step_runner, step: StepRun):
        super().__init__(step_runner, step)
        self.step = step
        self._docker_client = buildrunner.docker.new_client(
            timeout=step_runner.build_runner.docker_timeout,
        )
        if self.step_runner.network_name and not self._docker_client.networks(
            names=[self.step_runner.network_name]
        ):
            self.step_runner.log.info(
                f"Creating network {self.step_runner.network_name}"
            )
            self._docker_client.create_network(
                name=self.step_runner.network_name,
                labels=self.step_runner.container_labels,
            )
        self._source_container = None
        self._service_runners = OrderedDict()
        self._service_links = {}
        self._sshagent = None
        self._dockerdaemonproxy = None
        self.runner = None
        self.images_to_remove = []

    def __del__(self):
        if self.step_runner.network_name and self._docker_client.networks(
            names=[self.step_runner.network_name]
        ):
            self.step_runner.log.info(
                f"Deleting network {self.step_runner.network_name}"
            )
            try:
                self._docker_client.remove_network(self.step_runner.network_name)
            except Exception as _ex:
                self.step_runner.log.error(f"Error removing network: {_ex}")

    def _get_source_container(self):
        """
        Get (creating the container if necessary) the container id of the
        source container.
        """
        if not self._source_container:
            self._source_container = self._docker_client.create_container(
                self.step_runner.build_runner.get_source_image(),
                command="/bin/sh",
                labels=self.step_runner.container_labels,
                networking_config=self._docker_client.create_networking_config({
                    self.step_runner.network_name: self._docker_client.create_endpoint_config()
                })
                if self.step_runner.network_name
                else None,
            )["Id"]
            self._docker_client.start(
                self._source_container,
            )
            self.step_runner.log.info(
                f"Created source container {self._source_container:.10}"
            )
        return self._source_container

    def _process_volumes_from(self, volumes_from):
        """
        Translate the volumes_from configuration to the appropriate service
        container ids.
        """
        _volumes_from = []
        for sc_vf in volumes_from:
            volumes_from_definition = sc_vf.rsplit(":")
            service_container = volumes_from_definition[0]
            volume_option = None
            if len(volumes_from_definition) > 1:
                volume_option = volumes_from_definition[1]
            if service_container not in list(self._service_links.values()):
                raise BuildRunnerConfigurationError(
                    f'"volumes_from" configuration "{sc_vf}" does not reference a valid service container\n'
                )
            for container, service in self._service_links.items():
                if service == service_container:
                    if volume_option:
                        _volumes_from.append(
                            f"{container}:{volume_option}",
                        )
                    else:
                        _volumes_from.append(container)
                    break
        return _volumes_from

    def _retrieve_artifacts(self, console=None):  # pylint: disable=too-many-locals
        """
        Gather artifacts from the build container and place in the
        step-specific results dir.

        This function actually spins up a separate docker container that mounts
        the build container's source directory, queries for artifact patterns,
        then copies to the step-specific results directory (which is mounted as
        a host volume). It also appropriately sets file ownership of the
        archived artifacts to the user running the buildrunner process.
        """
        if not self.step.artifacts:
            return
        self.step_runner.log.info("Gathering artifacts")

        # use a small busybox image to list the files matching the glob
        artifact_lister = None
        try:
            image_config = DockerRunner.ImageConfig(
                f"{BuildRunnerConfig.get_instance().global_config.docker_registry}/{self.ARTIFACT_LISTER_DOCKER_IMAGE}",
                pull_image=False,
            )
            artifact_lister = DockerRunner(
                image_config,
                log=self.step_runner.log,
                # Log run messages at a debug level
                run_log_debug=True,
            )
            # NOTE: see if we can use archive commands to eliminate the need for
            #       the /stepresults volume when we can move to api v1.20
            artifact_lister.start(
                volumes_from=[self._get_source_container()],
                volumes={
                    self.step_runner.results_dir: "/stepresults",
                },
                working_dir=SOURCE_VOLUME_MOUNT,
                shell="/bin/sh",
            )

            for pattern, properties in self.step.artifacts.items():
                # query files for each artifacts pattern, capturing the output
                # for parsing
                stat_output_file = f"{str(uuid.uuid4())}.out"
                stat_output_file_local = os.path.join(
                    self.step_runner.results_dir,
                    stat_output_file,
                )
                exit_code = artifact_lister.run(
                    f'stat -c "%n{FILE_INFO_DELIMITER}%F" {pattern} >/stepresults/{stat_output_file}',
                    console=console,
                    stream=True,
                    log=self.step_runner.log,
                )

                # if the command was successful we found something
                if exit_code == 0:
                    with open(
                        stat_output_file_local, "r", encoding="utf-8"
                    ) as output_fd:
                        output = output_fd.read()
                    artifact_files = [af.strip() for af in output.split("\n")]
                    for art_info in artifact_files:
                        if not art_info or FILE_INFO_DELIMITER not in art_info:
                            continue
                        artifact_file, file_type = art_info.split(
                            FILE_INFO_DELIMITER,
                        )

                        # check if the file is directory, to copy recursive
                        is_dir = file_type.strip() == "directory"

                        if properties and properties.get("rename"):
                            if "*" in pattern:
                                raise BuildRunnerConfigurationError(
                                    f"Rename is not supported with wildcard patterns. `{pattern}` is not a valid pattern to use with rename."
                                )

                        if is_dir:
                            # directory => recursive copy
                            self._archive_dir(
                                artifact_lister,
                                properties,
                                artifact_file,
                            )

                        else:
                            output_file_name = os.path.basename(artifact_file)

                            if properties and properties.get("rename"):
                                output_file_name = properties.get("rename")
                                properties["rename"] = {
                                    "old": artifact_file,
                                    "new": properties.get("rename"),
                                }

                            new_artifact_file = "/stepresults/" + output_file_name
                            archive_command = (
                                "cp",
                                "-L",
                                artifact_file,
                                new_artifact_file,
                            )
                            self._archive_file(
                                artifact_lister,
                                "file",
                                output_file_name,
                                archive_command,
                                artifact_file,
                                new_artifact_file,
                                output_file_name,
                                properties,
                            )

                # remove the stat output file
                if os.path.exists(stat_output_file_local):
                    os.remove(stat_output_file_local)

        finally:
            if artifact_lister:
                # make sure the current user/group ids of our
                # process are set as the owner of the files
                exit_code = artifact_lister.run(
                    f"chown -R {int(os.getuid())}:{int(os.getgid())} /stepresults",
                    console=console,
                    log=self.step_runner.log,
                )
                if exit_code != 0:
                    # pylint: disable=broad-exception-raised
                    raise Exception(
                        "Error gathering artifacts--unable to change ownership"
                    )

                artifact_lister.cleanup()

    def _archive_dir(
        self,
        artifact_lister,
        properties,
        artifact_file,
    ):  # pylint: disable=too-many-locals
        """
        Archive the given directory.
        """
        # Rename doesn't apply to uncompressed directories
        if properties and properties.get("format", None) == "uncompressed":
            # recursively find all files in dir and add
            # each one, passing any properties
            find_output_file = f"{str(uuid.uuid4())}.out"
            find_output_file_local = os.path.join(
                self.step_runner.results_dir,
                find_output_file,
            )
            find_exit_code = artifact_lister.run(
                f"find {artifact_file} -type f >/stepresults/{find_output_file}",
                stream=False,
                log=self.step_runner.log,
            )
            if find_exit_code == 0:
                with open(find_output_file_local, "r", encoding="utf-8") as output_fd:
                    output = output_fd.read()
                for _file in [_f.strip() for _f in output.split("\n")]:
                    if not _file:
                        continue

                    output_file_name = _file
                    new_artifact_file = "/stepresults/" + output_file_name
                    _dir_name = os.path.dirname(_file)
                    archive_command = (
                        "mkdir",
                        "-p",
                        "/stepresults/" + _dir_name,
                    )
                    exit_code = artifact_lister.run(
                        archive_command,
                        log=self.step_runner.log,
                    )
                    if exit_code != 0:
                        # pylint: disable=broad-exception-raised
                        raise Exception(
                            f"Error gathering artifact {artifact_file}",
                        )
                    archive_command = (
                        "cp",
                        "-r",
                        _file,
                        new_artifact_file,
                    )
                    self._archive_file(
                        artifact_lister,
                        "file",
                        _file,
                        archive_command,
                        _file,
                        new_artifact_file,
                        output_file_name,
                        properties,
                    )

                # remove the find output file
                if os.path.exists(find_output_file_local):
                    os.remove(find_output_file_local)

        else:
            filename = os.path.basename(artifact_file)

            dest_filename = filename
            if properties and properties.get("rename"):
                dest_filename = properties.get("rename")
                properties["rename"] = {
                    "old": artifact_file,
                    "new": properties.get("rename"),
                }

            arch_props = {
                "name": dest_filename,
                "compression": "gz",
                "type": "tar",
            }
            if properties:
                arch_props.update(properties.get("archive", {}))

            workdir = None
            if arch_props["type"] == "tar":
                suffix = arch_props.get(
                    "suffix", f".{arch_props['type']}.{arch_props['compression']}"
                )
                output_file_name = arch_props["name"] + suffix
                new_artifact_file = "/stepresults/" + output_file_name
                archive_command = [
                    "tar",
                    self.TAR_COMPRESSION_ARG.get(
                        arch_props["compression"], "--auto-compress"
                    ),
                    "--xform",
                    f"s|^{filename}|{arch_props['name']}|",
                    "-cv",
                ]
                if os.path.dirname(artifact_file):
                    archive_command.extend((
                        "-C",
                        os.path.dirname(artifact_file),
                    ))
                archive_command.extend((
                    "-f",
                    new_artifact_file,
                    filename,
                ))

            elif arch_props["type"] == "zip":
                output_file_name = f"{arch_props['name']}.{arch_props['type']}"
                new_artifact_file = "/stepresults/" + output_file_name
                archive_command = [
                    "zip",
                    new_artifact_file,
                    artifact_file,
                ]

            new_properties = {}
            if properties:
                new_properties.update(properties)
            new_properties["buildrunner.compressed.directory"] = "true"
            self._archive_file(
                artifact_lister,
                "directory",
                filename,
                archive_command,
                artifact_file,
                new_artifact_file,
                output_file_name,
                new_properties,
                workdir=workdir,
            )

    def _archive_file(
        self,
        artifact_lister,
        file_type,
        filename,
        archive_command,
        artifact_file,
        new_artifact_file,
        output_file_name,
        properties,
        workdir=None,
    ):  # pylint: disable=too-many-arguments
        """
        Archive the given file.
        """
        # Unused arg
        _ = new_artifact_file

        self.step_runner.log.debug(f"- found {file_type} {filename}")

        exit_code = artifact_lister.run(
            archive_command,
            log=self.step_runner.log,
            workdir=workdir,
        )
        if exit_code != 0:
            # pylint: disable=broad-exception-raised
            raise Exception(
                f"Error gathering artifact {artifact_file}",
            )

        if not properties or (
            isinstance(properties, dict) and properties.get("push", True)
        ):
            # register the artifact with the run controller
            self.step_runner.build_runner.add_artifact(
                os.path.join(
                    self.step_runner.name,
                    output_file_name,
                ),
                properties,
            )

    def _get_local_file(
        self, buildrunner_config: BuildRunnerConfig, file_alias: str
    ) -> str:
        """
        Lookup file alias from global config or as a local file.
        :param buildrunner_config: buildrunner global configuration
        :param file_alias: the file alias
        :return: the local file to use (if available)
        :raises BuildRunnerConfigurationError: the file alias is not found and is not a valid local file
        """
        f_local = buildrunner_config.get_local_files_from_alias(file_alias)
        if not f_local:
            f_local = os.path.realpath(
                os.path.join(self.step_runner.build_runner.build_dir, file_alias)
            )
            if (
                f_local != self.step_runner.build_runner.build_dir
                and not f_local.startswith(
                    self.step_runner.build_runner.build_dir + os.path.sep
                )
            ):
                raise BuildRunnerConfigurationError(
                    f'Mount path of "{file_alias}" attempts to step out of '
                    f'source directory "{self.step_runner.build_runner.build_dir}"'
                )

            if not os.path.exists(f_local):
                raise BuildRunnerConfigurationError(
                    f"Cannot find valid alias for files entry '{file_alias}' nor path at '{f_local}'"
                )
        return f_local

    # pylint: disable=too-many-statements,too-many-branches,too-many-locals
    def _start_service_container(self, name, service: Service):
        """
        Start a service container.
        """
        buildrunner_config = BuildRunnerConfig.get_instance()
        _image = None
        # see if we need to build an image
        if service.build:
            build_image_task = BuildBuildStepRunnerTask(
                self.step_runner,
                service.build,
            )
            _build_context = {}
            build_image_task.run(_build_context)
            _image = _build_context.get("image", None)

        if service.image:
            _image = service.image
        assert _image

        self.step_runner.log.info(
            f'Creating service container "{name}" from image "{_image}"'
        )
        service_logger = ContainerLogger.for_service_container(name)

        # setup custom env variables
        _env = dict(buildrunner_config.env)
        _env["BUILDRUNNER_INVOKE_USER"] = pwd.getpwuid(os.getuid())
        _env["BUILDRUNNER_INVOKE_UID"] = os.getuid()
        _env["BUILDRUNNER_INVOKE_GROUP"] = grp.getgrgid(os.getgid())
        _env["BUILDRUNNER_INVOKE_GID"] = os.getgid()

        # do we need to change to a given dir when running
        # a cmd or script?
        _cwd = service.cwd

        # need to expose any ports?
        _ports = service.ports

        # default to a container that runs the default
        # image command
        _shell = None

        # do we need to run an explicit cmd?
        if service.cmd:
            # if so we need to run the cmd within a default
            # shell--specify it here
            _shell = DEFAULT_SHELL

        # if a shell is specified use it
        if service.shell:
            _shell = service.shell

        # see if there are any provisioners defined
        _provisioners = None
        if service.provisioners:
            _provisioners = create_provisioners(
                service.provisioners,
                service_logger,
            )

        # determine if a user is specified
        _user = service.user

        # determine if a hostname is specified
        _hostname = service.hostname or name

        # determine if a dns host is specified
        _dns = None
        if service.dns:
            # If the dns host is set to a string and that string is a reference to a running service
            # container, pull the ip address out of the service container.
            _dns = self._resolve_service_ip(service.dns)

        # determine if a dns_search domain is specified
        _dns_search = service.dns_search

        # determine if any extra hosts were provided
        _extra_hosts = None
        if service.extra_hosts:
            _extra_hosts = {
                extra_host: self._resolve_service_ip(extra_host_ip)
                for extra_host, extra_host_ip in service.extra_hosts.items()
            }

        # set service specific environment variables
        _env["BUILDRUNNER_STEP_ID"] = self.step_runner.id
        _env["BUILDRUNNER_STEP_NAME"] = self.step_runner.name
        if service.env:
            for key, value in service.env.items():
                _env[key] = value

        # make buildrunner aware of spawned containers
        _containers = service.containers

        # wait for ports on this container to be listening
        # before moving on
        _wait_for = []
        if service.wait_for:
            _wait_for = service.wait_for

        _volumes_from = [self._get_source_container()]

        # attach the ssh agent to the service container
        if self._sshagent and service.inject_ssh_agent:
            ssh_container, ssh_env = self._sshagent.get_info()
            if ssh_container:
                _volumes_from.append(ssh_container)
            if ssh_env:
                for _var, _val in ssh_env.items():
                    _env[_var] = _val

        # attach the docker daemon container
        daemon_container, daemon_env = self._dockerdaemonproxy.get_info()
        if daemon_container:
            _volumes_from.append(daemon_container)
        if daemon_env:
            for _var, _val in daemon_env.items():
                _env[_var] = _val

        # see if we need to map any service container volumes
        if service.volumes_from:
            _volumes_from.extend(
                self._process_volumes_from(
                    service.volumes_from,
                )
            )

        _volumes = {
            self.step_runner.build_runner.build_results_dir: ARTIFACTS_VOLUME_MOUNT
            + ":ro",
        }
        if service.files:
            for f_alias, f_path in service.files.items():
                # lookup file from alias
                f_local = self._get_local_file(buildrunner_config, f_alias)

                if f_path[-3:] not in [":ro", ":rw"]:
                    f_path = f_path + ":ro"

                _volumes[f_local] = f_path

                service_logger.info(f"Mounting {f_local} -> {f_path}")

        # instantiate and start the runner
        image_config = DockerRunner.ImageConfig(
            _image,
            self.step.pull if service.pull is not None else True,
            self.step.platform,
        )
        service_runner = DockerRunner(
            image_config,
            log=service_logger,
        )
        self._service_runners[name] = service_runner
        cont_name = self.step_runner.id + "-" + name
        systemd = self.is_systemd(service, _image)
        service_container_id = service_runner.start(
            name=cont_name,
            volumes=_volumes,
            volumes_from=_volumes_from,
            ports=_ports,
            links=self._service_links,
            shell=_shell,
            provisioners=_provisioners,
            environment=_env,
            user=_user,
            hostname=_hostname,
            dns=_dns,
            dns_search=_dns_search,
            extra_hosts=_extra_hosts,
            working_dir=_cwd,
            containers=_containers,
            systemd=systemd,
            systemd_cgroup2=self.is_systemd_cgroup2(systemd, service, _image),
            network=self.step_runner.network_name,
        )
        self._service_links[cont_name] = name

        def attach_to_service():
            """Function to attach to service in a separate thread."""
            # if specified, run a command
            if service.cmd:
                exit_code = service_runner.run(
                    service.cmd,
                    console=service_logger,
                    # log=self.step_runner.log,
                )
                if exit_code != 0:
                    service_logger.info(
                        f'Service command "{service.cmd}" exited with code {exit_code}'
                    )
            else:
                service_runner.attach_until_finished(service_logger)
            service_logger.cleanup()

        # Attach to the container in a separate thread
        service_management_thread = threading.Thread(
            name=f"{self.step_runner.name}--{name}",
            target=attach_to_service,
        )
        service_management_thread.daemon = True
        service_management_thread.start()

        # wait for listening ports on this container
        for wait_for_data in _wait_for:
            self.wait(cont_name, wait_for_data)

        self.step_runner.log.info(
            f'Started service container "{name}" ({service_container_id:.10})'
        )

    def wait(self, name, wait_for_data):
        """
        Wait for listening port on named container
        """
        network_settings = self._docker_client.inspect_container(name).get(
            "NetworkSettings", {}
        )
        if self.step_runner.network_name:
            ipaddr = (
                network_settings.get("Networks", {})
                .get(self.step_runner.network_name, {})
                .get("IPAddress", None)
            )
        else:
            ipaddr = network_settings.get("IPAddress", None)
        socket_open = False

        if isinstance(wait_for_data, dict):
            port = wait_for_data.get("port")
            if not port:
                raise BuildRunnerProcessingError(
                    "Configuration error, wait_for must include a port"
                )
            timeout = wait_for_data.get("timeout", self.WAIT_FOR_DEFAULT_TIMEOUT)
            if not isinstance(timeout, int):
                timeout = int(timeout)
        else:
            port = wait_for_data
            timeout = self.WAIT_FOR_DEFAULT_TIMEOUT

        start_time = time.time()
        while not socket_open:
            time_spent = int(time.time() - start_time)
            self.step_runner.log.info(
                f"Waiting for port {port} to be listening for connections in container {name} "
                f"with IP address {ipaddr} ({time_spent}/{timeout} seconds elapsed)"
            )

            # check that the container is still available
            container = self._docker_client.inspect_container(name)
            container_status = container.get("State", {}).get("Status")
            if container_status not in ["created", "running"]:
                raise BuildRunnerProcessingError(
                    f"Unable to wait for a service port {port} to be ready, the container"
                    f" {name} status is {container_status}"
                )

            # Use a small nc image to test if the port is open from within the docker network
            # Linux can talk to containers directly, but mac and other OSes cannot
            # See https://github.com/docker/for-mac/issues/155 for more info for mac
            nc_tester = None
            try:
                image_config = DockerRunner.ImageConfig(
                    f"{BuildRunnerConfig.get_instance().global_config.docker_registry}/{self.NC_DOCKER_IMAGE}",
                    pull_image=False,
                )
                nc_tester = DockerRunner(
                    image_config,
                    # Do not log anything from this container
                    log=None,
                )
                nc_tester.start(
                    # The shell is the command
                    shell=f"-n -z {ipaddr} {port}",
                    network=self.step_runner.network_name,
                )

                nc_tester.attach_until_finished()
                exit_code = nc_tester.exit_code
                socket_open = exit_code is not None and not exit_code
            finally:
                if nc_tester:
                    nc_tester.cleanup()

            if not socket_open:
                # Make sure we have not yet timed out
                if time_spent > timeout:
                    raise BuildRunnerProcessingError(
                        f"Timed out waiting for port {port} to be opened in container {name} "
                        f"with IP address {ipaddr} after {timeout} seconds"
                    )
                time.sleep(1)

        self.step_runner.log.info(
            f"Port {int(port)} is listening in container {name} with IP address {ipaddr}"
        )

    def _resolve_service_ip(self, service_name):
        """
        If service_name represents a running service, return its IP address.
        Otherwise, return the service_name
        """
        rval = service_name
        if isinstance(service_name, str) and service_name in self._service_runners:
            ipaddr = self._service_runners[service_name].get_ip(
                self.step_runner.network_name
            )
            if ipaddr is not None:
                rval = ipaddr
        return rval

    def run(self, context: dict):  # pylint: disable=too-many-statements,too-many-branches,too-many-locals
        buildrunner_config = BuildRunnerConfig.get_instance()

        # Set pull attribute based on configured image
        # Default to configuration
        if self.step.pull is not None:
            pull_image = self.step.pull
            self.step_runner.log.info(
                f"Pulling image was overridden via config to {pull_image}"
            )
        else:
            # Default to pulling, but check if image was committed by buildrunner and set to false if so
            pull_image = True
            config_image = self.step.image
            if config_image:
                pull_image = (
                    config_image not in self.step_runner.build_runner.committed_images
                )
                self.step_runner.log.info(
                    f"Pull was not specified in configuration, defaulting to {pull_image}"
                )

        _run_image = self.step.image
        if not _run_image:
            _run_image = context.get("image")
        if not _run_image:
            raise BuildRunnerConfigurationError(
                'Docker run context must specify a "image" attribute or be preceded by a build context'
            )
        _lrun_image = _run_image.lower()
        if _lrun_image != _run_image:
            self.step_runner.log.info(
                f"Forcing image name to lowercase: {_run_image} => {_lrun_image}"
            )
            _run_image = _lrun_image

        self.step_runner.log.info(f'Creating build container from image "{_run_image}"')
        container_logger = ContainerLogger.for_build_container(self.step_runner.name)
        container_meta_logger = ContainerLogger.for_build_container(
            self.step_runner.name
        )

        # container defaults
        _source_container = self._get_source_container()
        _container_name = str(uuid.uuid4())
        _env_defaults = dict(buildrunner_config.env)
        _env_defaults.update({
            "BUILDRUNNER_SOURCE_CONTAINER": _source_container,
            "BUILDRUNNER_BUILD_CONTAINER": _container_name,
        })
        container_args = {
            "name": _container_name,
            "hostname": None,
            "working_dir": SOURCE_VOLUME_MOUNT,
            "shell": None,
            "user": None,
            "provisioners": None,
            "dns": None,
            "dns_search": None,
            "extra_hosts": None,
            "environment": _env_defaults,
            "containers": None,
            "ports": None,
            "volumes_from": [_source_container],
            "volumes": {
                self.step_runner.build_runner.build_results_dir: (
                    ARTIFACTS_VOLUME_MOUNT + ":ro"
                ),
            },
            "cap_add": None,
            "privileged": None,
        }
        caches = OrderedDict()

        # see if we need to inject ssh keys
        if self.step.ssh_keys:
            self._sshagent = DockerSSHAgentProxy(
                self._docker_client,
                self.step_runner.log,
                buildrunner_config.global_config.docker_registry,
                self.step_runner.multi_platform,
                self.step_runner.container_labels,
                self.step_runner.network_name,
            )
            self._sshagent.start(
                buildrunner_config.get_ssh_keys_from_aliases(
                    self.step.ssh_keys,
                )
            )

        # start the docker daemon proxy
        self._dockerdaemonproxy = DockerDaemonProxy(
            self._docker_client,
            self.step_runner.log,
            buildrunner_config.global_config.docker_registry,
            self.step_runner.container_labels,
            self.step_runner.network_name,
        )
        self._dockerdaemonproxy.start()

        # start any service containers
        if self.step.services:
            for _name, _service in self.step.services.items():
                self._start_service_container(_name, _service)

        # determine if there is a command to run
        _cmds = []
        if self.step.cmd:
            container_args["shell"] = DEFAULT_SHELL
            _cmds.append(self.step.cmd)
        if self.step.cmds:
            container_args["shell"] = DEFAULT_SHELL
            _cmds.extend(self.step.cmds)

        if self.step.provisioners:
            container_args["shell"] = DEFAULT_SHELL
            container_args["provisioners"] = create_provisioners(
                self.step.provisioners,
                container_logger,
            )

        # if a shell is specified use it
        if self.step.shell:
            container_args["shell"] = self.step.shell

        # determine the working dir the build should be run in
        if self.step.cwd:
            container_args["working_dir"] = self.step.cwd

        # determine if a user is specified
        if self.step.user:
            container_args["user"] = self.step.user

        # determine if a hostname is specified
        if self.step.hostname:
            container_args["hostname"] = self.step.hostname

        # determine if a dns host is specified
        if self.step.dns:
            # If the dns host is set to a string and that string is a reference to a running service
            # container, pull the ip address out of the service container.
            container_args["dns"] = self._resolve_service_ip(self.step.dns)

        # determine if a dns_search domain is specified
        if self.step.dns_search:
            container_args["dns_search"] = self.step.dns_search

        # determine additional hosts to add
        if self.step.extra_hosts:
            extra_hosts = {}
            for extra_host, extra_host_ip in self.step.extra_hosts.items():
                extra_hosts[extra_host] = self._resolve_service_ip(extra_host_ip)
            container_args["extra_hosts"] = extra_hosts

        # set step specific environment variables
        container_args["environment"]["BUILDRUNNER_STEP_ID"] = self.step_runner.id
        container_args["environment"]["BUILDRUNNER_STEP_NAME"] = self.step_runner.name
        if self.step.env:
            for key, value in self.step.env.items():
                container_args["environment"][key] = value

        # make buildrunner aware of spawned containers
        if self.step.containers:
            container_args["containers"] = self.step.containers

        # see if we need to map any service container volumes
        if self.step.volumes_from:
            container_args["volumes_from"].extend(
                self._process_volumes_from(
                    self.step.volumes_from,
                )
            )

        # see if we need to attach to a sshagent container
        if self._sshagent:
            ssh_container, ssh_env = self._sshagent.get_info()
            if ssh_container:
                container_args["volumes_from"].append(ssh_container)
            if ssh_env:
                for _var, _val in ssh_env.items():
                    container_args["environment"][_var] = _val

        # attach the docker daemon container
        daemon_container, daemon_env = self._dockerdaemonproxy.get_info()
        if daemon_container:
            container_args["volumes_from"].append(daemon_container)
        if daemon_env:
            for _var, _val in daemon_env.items():
                container_args["environment"][_var] = _val

        # see if we need to inject any files
        if self.step.files:
            for f_alias, f_path in self.step.files.items():
                f_local = self._get_local_file(buildrunner_config, f_alias)

                if f_path[-3:] not in [":ro", ":rw"]:
                    f_path = f_path + ":ro"

                container_args["volumes"][f_local] = f_path

                container_meta_logger.info(f"Mounting {f_local} -> {f_path}")

        if self.step.caches:
            for key, value in self.step.caches.items():
                if isinstance(value, str):
                    # get the cache location from the main BuildRunner class
                    cache_archive_file = (
                        self.step_runner.build_runner.get_cache_archive_file(
                            cache_name=key,
                            project_name=buildrunner_config.vcs.name,
                        )
                    )
                    caches[cache_archive_file] = value
                    container_meta_logger.info(
                        f"Considering local cache `{cache_archive_file}` -> docker path `{value}`"
                    )
                elif isinstance(value, list):
                    for cache_local in value:
                        cache_archive_file = (
                            self.step_runner.build_runner.get_cache_archive_file(
                                cache_local,
                            )
                        )
                        caches[cache_archive_file] = key
                        container_meta_logger.info(
                            f"Considering local cache [{cache_archive_file}] -> docker path [{key}]"
                        )
                else:
                    container_meta_logger.warning(
                        f"Type {type(value)} is not supported. "
                        f"Not able to use cache functionality for {key}: {value}"
                    )
                    continue

        # add any capabilities when the container runs
        if self.step.cap_add:
            container_args["cap_add"] = self.step.cap_add

        # allow privileged containers (used sparingly)
        if self.step.privileged:
            container_args["privileged"] = self.step.privileged

        # only expose ports for a run step if the flag is set
        if self.step_runner.build_runner.publish_ports and self.step.ports:
            container_args["ports"] = self.step.ports

        exit_code = None
        try:
            # create and start runner, linking any service containers
            image_config = DockerRunner.ImageConfig(
                _run_image,
                pull_image=pull_image,
                platform=self.step.platform,
            )
            self.runner = DockerRunner(
                image_config,
                log=self.step_runner.log,
            )
            # Figure out if we should be running systemd.  Has to happen after docker pull
            container_args["systemd"] = self.is_systemd(self.step, _run_image)
            container_args["systemd_cgroup2"] = self.is_systemd_cgroup2(
                container_args["systemd"], self.step, _run_image
            )
            container_args["network"] = self.step_runner.network_name

            container_id = self.runner.start(
                links=self._service_links, **container_args
            )

            self.runner.restore_caches(container_meta_logger, caches)

            self.step_runner.log.info(f"Started build container {container_id:.10}")

            if _cmds:
                # First ensure that the source directory is marked as safe for newer git versions
                # Any errors are ignored here
                self.runner.run("git config --global --add safe.directory /source")

                # run each cmd
                for _cmd in _cmds:
                    container_meta_logger.info(f"cmd> {_cmd}")
                    exit_code = self.runner.run(
                        _cmd,
                        console=container_logger,
                        # log=self.step_runner.log,
                    )
                    if exit_code:
                        log_method = container_meta_logger.error
                    else:
                        log_method = container_meta_logger.info
                    log_method(f'Command "{_cmd}" exited with code {exit_code}')

                    if exit_code:
                        break
            else:
                self.runner.attach_until_finished(container_logger)
                exit_code = self.runner.exit_code
                container_meta_logger.info(f"Container exited with code {exit_code}")

            if exit_code:
                container_meta_logger.info(
                    "Skipping cache save due to failed exit code"
                )
            else:
                try:
                    self.runner.save_caches(
                        container_meta_logger, caches, container_args.get("environment")
                    )
                except Exception as _ex:
                    container_meta_logger.error(
                        f"Error saving caches, ignoring: {_ex}",
                    )
                    # Failing to save caches should not fail the build, just continue as normal

        finally:
            if self.runner:
                self.runner.stop()
            if container_logger:
                container_logger.cleanup()
            if container_meta_logger:
                container_meta_logger.cleanup()

        # gather artifacts to results dir even if run has bad exit code
        self._retrieve_artifacts(container_logger)

        # if we have an unsuccessful exit code abort
        if exit_code != 0:
            raise BuildRunnerProcessingError("Error running build container")

        context["run_runner"] = self.runner

        if self.step.post_build:
            self._run_post_build(context)

    def _run_post_build(self, context):
        """
        Commit the run image and perform a docker build, prepending the run
        image to the Dockerfile.
        """
        self.step_runner.log.info("Running post-build processing")
        post_build = self.step.post_build
        temp_tag = f"buildrunner-post-build-tag-{str(uuid.uuid4())}"
        python_on_whales.docker.image.tag(
            self.runner.commit(self.step_runner.log),
            temp_tag,
        )
        self.images_to_remove.append(temp_tag)

        post_build.pull = False
        build_image_task = BuildBuildStepRunnerTask(
            self.step_runner,
            post_build,
            image_to_prepend_to_dockerfile=temp_tag,
            force_legacy_builder=True,
        )
        _build_context = {}
        build_image_task.run(_build_context)
        context["run-image"] = _build_context.get("image", None)

    def cleanup(self, context):  # pylint: disable=unused-argument
        if self.runner:
            if self.runner.container:
                self.step_runner.log.info(
                    f"Destroying build container {self.runner.container['Id']:.10}"
                )
            self.runner.cleanup()

        if self._service_runners:
            for _sname, _srun in reversed(list(self._service_runners.items())):
                self.step_runner.log.info(f'Destroying service container "{_sname}"')
                _srun.cleanup()

        if self._dockerdaemonproxy:
            self._dockerdaemonproxy.stop()

        if self._sshagent:
            self._sshagent.stop()

        if self._source_container:
            self.step_runner.log.info(
                f"Destroying source container {self._source_container:.10}"
            )
            self._docker_client.remove_container(
                self._source_container,
                force=True,
                v=True,
            )

        for image in self.images_to_remove:
            python_on_whales.docker.image.remove(image, force=True)

    def _get_label_is_truthy(self, image, label_name: str) -> bool:
        labels = (
            self._docker_client.inspect_image(image).get("Config", {}).get("Labels", {})
        )
        if not labels or label_name not in labels:
            return False
        # Labels will be set as the string value.  Make sure we handle '0' and 'False'
        value = labels.get(label_name, False)
        return bool(value and value != "0" and value != "False")

    def is_systemd(self, run_service: RunAndServicesBase, image: str) -> bool:
        """
        Check if an image runs systemd
        """
        if run_service.systemd is not None:
            return run_service.systemd
        return self._get_label_is_truthy(image, "BUILDRUNNER_SYSTEMD")

    def is_systemd_cgroup2(
        self, systemd: bool, run_service: RunAndServicesBase, image: str
    ) -> bool:
        """
        Check if an image needs the changes for cgroup2
        """
        if not systemd:
            # Do not run any other checks if we are not using systemd at all
            return False

        if run_service.systemd_cgroup2 is not None:
            return run_service.systemd_cgroup2
        return self._get_label_is_truthy(image, "BUILDRUNNER_SYSTEMD_CGROUP2")



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

from typing import Optional

from ..models import GlobalConfig


def fetch_file(parsed_url, config: Optional[GlobalConfig]):
    """
    Fetch files using HTTP.
    """

    raise NotImplementedError("Not Implemented: fetch.http.file_fetch")



"""
Copyright 2023 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import logging
import platform as python_platform
from typing import Dict, List, Optional

from pydantic import BaseModel


LOGGER = logging.getLogger(__name__)

# Maps from lower-case platform.system() result to platform prefix in docker
PLATFORM_SYSTEM_MAPPINGS = {"darwin": "linux"}
PLATFORM_MACHINE_MAPPINGS = {
    "aarch64": "arm",
    "x86_64": "amd",
}


class TaggedImageInfo(BaseModel):
    # The image repo with no tag (e.g. repo/image-name)
    repo: str
    # All tags for this image
    tags: List[str]

    @property
    def image_refs(self) -> List[str]:
        """Returns a list of image names with tags"""
        return [f"{self.repo}:{tag}" for tag in self.tags]

    def __str__(self) -> str:
        return f"{self.repo}:{','.join(self.tags)}"


class BuiltTaggedImage(BaseModel):
    # The image repo with no tag (e.g. repo/image-name)
    repo: str
    # The single tag for this image
    tag: str
    # Digest for the image
    digest: str
    # Platform for the image
    platform: str

    @property
    def trunc_digest(self) -> str:
        return self.digest.replace("sha256:", "")[:12]

    @property
    def image_ref(self) -> str:
        """Returns an image name with the tag"""
        return f"{self.repo}:{self.tag}"


class BuiltImageInfo(BaseModel):
    """
    Contains information about images created during the build,
    including the final tagged image name and any tags (once generated).
    """

    # A unique ID for this built image
    id: str
    # The built images by each platform
    images_by_platform: Dict[str, BuiltTaggedImage] = {}
    # The images that should be created from the built images referenced in each instance
    tagged_images: List[TaggedImageInfo] = []

    @property
    def platforms(self) -> List[str]:
        return list(self.images_by_platform.keys())

    @property
    def built_images(self) -> List[BuiltTaggedImage]:
        return list(self.images_by_platform.values())

    def image_for_platform(self, platform: str) -> Optional[BuiltTaggedImage]:
        """Retrieves the built image for the given platform."""
        return self.images_by_platform.get(platform)

    def add_platform_image(self, platform: str, image_info: BuiltTaggedImage) -> None:
        if platform in self.images_by_platform:
            raise ValueError(
                f"Cannot add image {image_info} for platform {platform} since it already exists in this built image info"
            )
        self.images_by_platform[platform] = image_info

    def add_tagged_image(self, repo: str, tags: List[str]) -> TaggedImageInfo:
        """
        Creates a tagged image instance, adds it to the final image tags, then returns it.
        :arg repo: The image repo without the tag
        :arg tags: The tags to use when tagging the image
        """
        tagged_image = TaggedImageInfo(
            repo=repo,
            tags=tags,
        )
        self.tagged_images.append(tagged_image)
        return tagged_image

    @property
    def native_platform_image(self) -> BuiltTaggedImage:
        """
        Returns the built native platform image(s) matching the current machine's platform or the first built
        image if none matches.

        Args:
            name (str): The name of the image to find

        Returns:
            str: The name of the native platform image
        """
        # Converts the python platform results to docker-equivalents, and then matches based on the prefix
        # e.g. Darwin/aarch64 would become linux/arm which would then match linux/arm64/v7, etc
        native_system = python_platform.system().lower()
        native_machine = python_platform.machine()
        native_system = PLATFORM_SYSTEM_MAPPINGS.get(native_system, native_system)
        native_machine = PLATFORM_MACHINE_MAPPINGS.get(native_machine, native_machine)
        LOGGER.debug(
            f"Finding native platform for {self} for {native_system}/{native_machine}"
        )

        # Search for the image matching the native platform
        # Uses dash since this is matching on tag names which do not support forward slashes
        pattern = f"{native_system}-{native_machine}"
        matched_image = None
        for image in self.images_by_platform.values():
            if image.tag.replace(f"{self.id}-", "").startswith(pattern):
                matched_image = image

        # Still no matches found, get the first image and log a warning
        if not matched_image:
            platform = self.platforms[0]
            LOGGER.warning(
                f"Could not find image matching native platform {pattern} for {self}, using platform {platform} instead locally"
            )
            return self.images_by_platform[platform]

        return matched_image

    def __str__(self):
        num_platforms = len(self.platforms)
        if num_platforms == 0:
            return f"<{self.id}>"
        elif num_platforms == 1:
            return f"<{self.images_by_platform[self.platforms[0]]} ({self.id})>"
        image_strs = [
            f"{platform}: {image}"
            for platform, image in self.images_by_platform.items()
        ]
        return f"<{', '.join(image_strs)} ({self.id})>"

    def __repr__(self):
        return self.__str__()



"""
Copyright 2024 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

from typing import Any, Dict, List, Optional, Union

from pydantic import (
    BaseModel,
    BeforeValidator,
    Field,
    ValidationInfo,
    field_validator,
    model_validator,
)
from typing_extensions import Annotated


def _validate_artifact_type(value) -> Any:
    if value and not Artifact.model_validate(value):
        raise ValueError(f"Invalid artifact type: {value}")
    return value


AnnotatedArtifact = Annotated[Any, BeforeValidator(_validate_artifact_type)]


class StepPushSecurityScanConfig(BaseModel, extra="forbid"):
    enabled: Optional[bool] = None
    scanner: Optional[str] = None
    version: Optional[str] = None
    config: Optional[dict] = None
    max_score_threshold: Optional[float] = Field(None, alias="max-score-threshold")


class StepTask(BaseModel, extra="forbid"):
    """
    Used for type checking.
    """


class StepPypiPush(StepTask):
    """Step pypi push model"""

    repository: str
    username: Optional[str] = None
    password: Optional[str] = None
    skip_existing: bool = False


class Artifact(BaseModel):
    """Artifact model"""

    # Intentionally loose restrictions
    format: Optional[str] = None
    type: Optional[Any] = None
    compression: Optional[str] = None
    push: Optional[bool] = None
    rename: Optional[str] = None


class StepBuild(StepTask):
    """Build model within a step"""

    path: Optional[str] = None
    dockerfile: Optional[str] = None
    target: Optional[str] = None
    pull: Optional[bool] = None
    platform: Optional[str] = None
    platforms: Optional[List[str]] = None
    inject: Optional[Dict[str, Optional[str]]] = None
    no_cache: Optional[bool] = Field(alias="no-cache", default=None)
    buildargs: Optional[Dict[str, Any]] = None
    cache_from: Optional[
        Union[str, Dict[str, str], List[Dict[str, str]], List[str]]
    ] = None
    cache_to: Optional[Union[str, Dict[str, str]]] = None
    # import is a python reserved keyword so we need to alias it
    import_param: Optional[str] = Field(alias="import", default=None)
    secrets: Optional[List[str]] = None


class RunAndServicesBase(StepTask):
    """
    Base model for Run and Service which has several common fields
    """

    image: Optional[str] = None
    cmd: Optional[str] = None
    # Intentionally loose restrictions
    provisioners: Optional[Dict[str, str]] = None
    shell: Optional[str] = None
    cwd: Optional[str] = None
    user: Optional[str] = None
    hostname: Optional[str] = None
    dns: Optional[List[str]] = None
    dns_search: Optional[str] = None
    extra_hosts: Optional[Dict[str, str]] = None
    env: Optional[Dict[str, Optional[Any]]] = None
    files: Optional[Dict[str, str]] = None
    volumes_from: Optional[List[str]] = None
    ports: Optional[Dict[int, Optional[int]]] = None
    pull: Optional[bool] = None
    systemd: Optional[bool] = None
    systemd_cgroup2: Optional[bool] = None
    containers: Optional[List[str]] = None
    caches: Optional[Dict[str, Union[str, List[str]]]] = None


class Service(RunAndServicesBase):
    build: Optional[StepBuild] = None
    wait_for: Optional[List[Any]] = None
    inject_ssh_agent: Optional[bool] = Field(alias="inject-ssh-agent", default=None)
    # Not sure if this is valid, but it is in a test file
    # Didn't use StepRun because of the potential to have a infinitely nested model
    run: Optional[Any] = None

    @field_validator("build", mode="before")
    @classmethod
    def transform_build(cls, val) -> Optional[dict]:
        if not isinstance(val, str):
            return val
        return {
            "path": val,
        }

    @model_validator(mode="after")
    def validate_image_or_build(self):
        if not self.build and not self.image:
            raise ValueError("Service must specify an image or docker build context")
        if self.build and self.image:
            raise ValueError(
                "Service must specify either an image or docker build context, not both"
            )
        return self


class StepRun(RunAndServicesBase):
    """Run model within a step"""

    xfail: Optional[bool] = None
    services: Optional[Dict[str, Service]] = None
    cmds: Optional[List[str]] = None
    ssh_keys: Optional[List[str]] = Field(alias="ssh-keys", default=None)
    artifacts: Optional[Dict[str, Optional[AnnotatedArtifact]]] = None
    platform: Optional[str] = None
    cap_add: Optional[List[str]] = None
    privileged: Optional[bool] = None
    post_build: Optional[StepBuild] = Field(alias="post-build", default=None)
    no_cache: Optional[bool] = Field(alias="no-cache", default=None)

    @field_validator("post_build", mode="before")
    @classmethod
    def transform_post_build(cls, val) -> Optional[dict]:
        if not isinstance(val, str):
            return val
        return {
            "path": val,
        }

    @field_validator("cap_add", mode="before")
    @classmethod
    def transform_cap_add(cls, val) -> Optional[List[str]]:
        if not isinstance(val, str):
            return val
        return [val]


class StepRemote(StepTask):
    """Remote model within a step"""

    host: str
    cmd: str
    artifacts: Optional[Dict[str, Optional[AnnotatedArtifact]]] = None


class StepPushCommit(StepTask):
    """Push model within a step"""

    repository: str
    add_build_tag: bool = True
    tags: Optional[List[str]] = Field(
        None,
        min_length=1,
    )
    push: bool
    security_scan: Optional[StepPushSecurityScanConfig] = Field(
        None, alias="security-scan"
    )


class Step(BaseModel, extra="forbid"):
    """Step model"""

    # A build specified as a string is handled in the field validator
    build: Optional[StepBuild] = None
    push: Optional[List[StepPushCommit]] = None
    commit: Optional[List[StepPushCommit]] = None
    remote: Optional[StepRemote] = None
    run: Optional[StepRun] = None
    depends: Optional[List[str]] = None
    pypi_push: Optional[List[StepPypiPush]] = Field(alias="pypi-push", default=None)

    @field_validator("build", mode="before")
    @classmethod
    def transform_build(cls, val) -> Optional[dict]:
        if not isinstance(val, str):
            return val
        return {
            "path": val,
        }

    @field_validator("pypi_push", mode="before")
    @classmethod
    def transform_pypi_push(cls, vals, info: ValidationInfo) -> Optional[List[dict]]:
        if not vals:
            return vals
        if not isinstance(vals, list):
            vals = [vals]
        for index, val in enumerate(vals):
            if not val:
                raise ValueError(f"{info.field_name}.{index} must be a valid value")
            if isinstance(val, str):
                vals[index] = {"repository": val}
        return vals

    @field_validator("commit", "push", mode="before")
    @classmethod
    def transform_commit_push(cls, vals, info: ValidationInfo) -> Optional[List[dict]]:
        if not vals:
            return vals
        if not isinstance(vals, list):
            vals = [vals]
        for index, val in enumerate(vals):
            if not val:
                raise ValueError(f"{info.field_name}.{index} must be a valid value")
            if isinstance(val, str):
                vals[index] = {"repository": val}
            # Set the push field dynamically based on the top level field name
            vals[index]["push"] = info.field_name == "push"
        return vals

    def is_multi_platform(self):
        """
        Check if the step is a multi-platform build step
        """
        return self.build and self.build.platforms is not None

    def has_secrets(self):
        """
        Check if the step has secrets
        """
        return self.build and self.build.secrets is not None



"""
Copyright 2024 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import codecs
import copy
import datetime
import re
from io import StringIO
from typing import Callable

import jinja2

from buildrunner.utils import load_config


def read_yaml_file(
    env: dict,
    log_generated_file_method: Callable[[bool, str, str], None],
    log_file: bool,
    filename: str,
):
    """
    Reads a file in the local workspace as Jinja-templated YAML and returns the contents.
    Throws an error on failure.
    """
    with codecs.open(filename, "r", encoding="utf-8") as _file:
        jtemplate = jinja2.Template(_file.read())
    context = copy.deepcopy(env)
    file_contents = jtemplate.render(context)
    log_generated_file_method(log_file, filename, file_contents)
    return load_config(StringIO(file_contents), filename)


def strftime(_build_time: int, _format="%Y-%m-%d", _ts=None):
    """
    Format the provided timestamp. If no timestamp is provided, build_time is used
    :param _build_time: The build timestamp. This is bound with functools.partial and is not required to pass in.
    :param _format: Format string - default "%Y-%m-%d"
    :param _ts: Timestamp to format - default self.build_time
    :return: Formatted date/time string
    """
    if _ts is None:
        _ts = _build_time
    _date = datetime.datetime.fromtimestamp(_ts)
    return _date.strftime(_format)


def raise_exception_jinja(message):
    """
    Raises an exception from a jinja template.
    """
    # pylint: disable=broad-exception-raised
    raise Exception(message)


def re_sub_filter(text, pattern, replace, count=0, flags=0):
    """
    Filter for regular expression replacement.
    :param text: The string being examined for ``pattern``
    :param pattern: The pattern to find in ``text``
    :param replace: The replacement for ``pattern``
    :param count: How many matches of ``pattern`` to replace with ``replace`` (0=all)
    :param flags: Regular expression flags
    """
    return re.sub(pattern, replace, text, count=count, flags=flags)


def re_split_filter(text, pattern, maxsplit=0, flags=0):
    """
    Filter for regular expression replacement.
    :param text: The string being examined for ``pattern``
    :param pattern: The pattern used to split ``text``
    :param maxsplit: How many instances of ``pattern`` to split (0=all)
    :param flags: Regular expression flags
    """
    return re.split(pattern, text, maxsplit=maxsplit, flags=flags)



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import os
import ssl
import tempfile
from typing import Tuple
import urllib.parse
import docker

from buildrunner.errors import BuildRunnerError, BuildRunnerConfigurationError


try:
    # Newer API
    Client = docker.client.Client  # pylint: disable=no-member
except Exception:  # pylint: disable=broad-except
    try:
        # Older API
        Client = docker.client.APIClient

    except Exception:  # pylint: disable=broad-except
        # Older API
        Client = docker.api.client.APIClient

DOCKER_API_VERSION = "auto"
DOCKER_DEFAULT_DOCKERD_URL = "unix:///var/run/docker.sock"
MAX_TIMEOUT = 3600  # 1 hour


class BuildRunnerContainerError(BuildRunnerError):
    """Error indicating an issue managing a Docker container"""

    pass


def new_client(
    dockerd_url=None,
    tls=False,
    tls_verify=False,
    cert_path=None,
    timeout=None,
):
    """
    Return a newly configured Docker client.
    """
    _dockerd_url = dockerd_url
    if not _dockerd_url:
        _dockerd_url = os.getenv("DOCKER_HOST", DOCKER_DEFAULT_DOCKERD_URL)

    _tls = tls

    tls_config = None
    if tls_verify or str(os.environ.get("DOCKER_TLS_VERIFY", "0")) == "1":
        _tls = True
        _cert_path = os.getenv("DOCKER_CERT_PATH", cert_path)
        if not _cert_path:
            raise BuildRunnerConfigurationError(
                "TLS connection specified but cannot determine cert path"
                " (from DOCKER_CERT_PATH env variable)"
            )

        ca_cert_path = os.path.join(_cert_path, "ca.pem")
        client_cert = (
            os.path.join(_cert_path, "cert.pem"),
            os.path.join(_cert_path, "key.pem"),
        )

        tls_config = docker.tls.TLSConfig(
            ssl_version=ssl.PROTOCOL_TLSv1,
            client_cert=client_cert,
            verify=ca_cert_path,
            assert_hostname=False,
        )

    if _tls:
        # make sure the scheme is https
        url_parts = urllib.parse.urlparse(_dockerd_url)
        if url_parts.scheme == "tcp":
            _dockerd_url = urllib.parse.urlunparse(("https",) + url_parts[1:])

    args = {}
    if timeout is not None:
        if timeout == 0:
            args["timeout"] = MAX_TIMEOUT
        else:
            args["timeout"] = timeout
    try:
        return Client(
            base_url=_dockerd_url, version=DOCKER_API_VERSION, tls=tls_config, **args
        )
    except docker.errors.DockerException as e:
        if "Connection aborted" in str(e) and "No such file or directory" in str(e):
            raise BuildRunnerContainerError(
                "Error connecting to Docker daemon. Please ensure the Docker daemon is running and try again."
            )
        else:
            raise BuildRunnerContainerError(f"An docker error occurred: {e}")


def force_remove_container(docker_client, container):
    """
    Force removes a container from the given docker client.
    :param docker_client: the docker client
    :param container: the container
    """
    docker_client.remove_container(
        container,
        force=True,
        v=True,
    )


def get_dockerfile(dockerfile: str, temp_dir: str = None) -> Tuple[str, bool]:
    """
    Check if the dockerfile exists, if not create a temporary file and write the dockerfile to it.
    :param dockerfile: the dockerfile
    :param temp_dir: the temporary directory
    :return: the dockerfile and a boolean indicating if the dockerfile was created
    """
    cleanup_dockerfile = False
    curr_dockerfile = None

    if dockerfile:
        if os.path.exists(dockerfile):
            curr_dockerfile = dockerfile
        else:
            # pylint: disable=consider-using-with
            df_file = tempfile.NamedTemporaryFile(delete=False, dir=temp_dir)
            try:
                df_file.write(dockerfile.encode("utf-8"))
                cleanup_dockerfile = True
                curr_dockerfile = df_file.name
            finally:
                df_file.close()
    return curr_dockerfile, cleanup_dockerfile



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import base64
import requests

from ..models import GlobalConfig, GithubModel
from ...errors import (
    BuildRunnerConfigurationError,
    BuildRunnerProtocolError,
)


def _clean_nones(items):
    return list(filter(None, items))


def _fetch_file(parsed_url, gh_config: GithubModel):
    """
    Fetch files using Github v3 protocol.
    """

    endpoint = gh_config.endpoint
    version = gh_config.version

    username = gh_config.username
    if not username:
        raise RuntimeError(f"Failed to look up username to access github {endpoint}")

    auth = (username, gh_config.app_token)
    url = "/".join(
        _clean_nones([
            endpoint,
            version,
            "users",
            username,
        ])
    )
    resp = requests.get(url, auth=auth, timeout=180)
    if resp.status_code != 200:
        raise BuildRunnerProtocolError(
            f"Failed authenticating {username} on {endpoint}"
        )

    if not parsed_url.path.startswith("/"):
        raise ValueError('URL must begin with "/"')

    fpath = parsed_url.path.split("/")
    ubuild = _clean_nones([endpoint, version, "repos", fpath[1], fpath[2], "contents"])
    ubuild.extend(fpath[3:])
    url = "/".join(ubuild)
    resp = requests.get(url, auth=auth, timeout=180)
    if resp.status_code != 200:
        raise BuildRunnerProtocolError(f"Failed fetching URL: {url}")
    json_c = resp.json()

    encoding = json_c.get("encoding")
    enc_contents = json_c.get("content")
    if encoding == "base64":
        dec_contents = base64.b64decode(enc_contents)
    else:
        raise NotImplementedError(
            f"No implementation to decode {encoding}-encoded contents"
        )

    if resp.encoding == "utf-8":
        contents = dec_contents.decode("utf-8")
    else:
        raise NotImplementedError(
            f"No implementation to decode {resp.encoding}-encoded contents"
        )

    return contents


def fetch_file(parsed_url, config: GlobalConfig):
    """
    Fetch a file from Github.
    """

    if parsed_url.scheme != "github":
        raise ValueError(f'URL scheme must be "github" but is "{parsed_url.github}"')

    ghcfg = config.github
    if not ghcfg:
        raise BuildRunnerConfigurationError(
            "Missing configuration for github in buildrunner.yaml"
        )

    nlcfg = ghcfg.get(parsed_url.netloc)
    if not nlcfg:
        gh_cfgs = ", ".join(ghcfg.keys())
        raise BuildRunnerConfigurationError(
            f"Missing github configuration for {parsed_url.netloc} in buildrunner.yaml"
            f" - known github configurations: {gh_cfgs}"
        )

    ver = nlcfg.version
    # NOTE: potentially the _fetch_file() works for other github API versions.
    if ver == "v3" or ver is None:
        contents = _fetch_file(parsed_url, nlcfg)
    else:
        raise NotImplementedError(f"No version support for github API version {ver}")

    return contents
